import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/*
GEBitWriter uses the 6 least significant bits (range 0x40
- 0x7F), because of problems with cross-platform compatibility
using the bytes 0x80 - 0xFF and line-break characters.
Specifically, some web browsers tend to convert them from
one character set to another.
*/


public class GEBitWriter
{
	public static final int			kMaxIntBits = 32;
	
	protected int					cBitsUsed;
	protected int					cValueMask;
	protected int					cBaseByte;
	
	protected PrintWriter			cOutputWriter;
	protected CharArrayWriter		cOutputArray;
	protected int					cBufferValue; // least significant bits are used
	protected int					cBufferSize; // size in bits
	protected int					cBitsWritten;
	
	
	public GEBitWriter(File data) throws IOException {
		cOutputWriter = new PrintWriter(new FileOutputStream(data));
		cOutputArray = new CharArrayWriter();
		
		cBufferValue = 0;
		cBufferSize = 0;
		cBitsWritten = 0;
		
		// uses only 6 least significant bits
		cBitsUsed = 6;
		cValueMask = 0x3F;
		cBaseByte = 0x40;
	}
	
	public GEBitWriter(String fileName) throws IOException {
		this(new File(fileName));
	}
	
	public void close() throws IOException {
		if (cOutputWriter == null)
			throw new IOException("GEBitWriter.close: no valid file to close");
		
		// "flush" buffer by filling up the last byte
		if (cBufferSize > 0)
			pushInt(cBitsUsed - cBufferSize, 0);
		
		cOutputArray.writeTo(cOutputWriter);
		cOutputWriter.close();
		cOutputWriter = null;
	}
	
	/**
	Note: Only six bits are written per byte.
	*/
	public int getBitsWritten() {
		return cBitsWritten;
	}
	
	/**
	Writes the least significant bits of "value" will only
	store positive values.
	*/
	public void pushInt(int bits, int value) throws IOException {
		if (cOutputWriter == null)
			throw new IOException("GEBitWriter.pushInt: no valid file to write to");
		
		if (bits < 1 || bits > kMaxIntBits)
			throw new IOException("GEBitWriter.pushInt: incorrect number of bits (" + bits + ") to push, min = 1, max = " + kMaxIntBits + "; bits written: " + getBitsWritten());
		
		if (bits > 16) {
			pushInt(bits - 16, value >> 16);
			bits = 16;
		}
		
		// push into buffer
		cBufferValue = cBufferValue << bits;
		cBufferSize = cBufferSize + bits;
		cBufferValue = cBufferValue + (value & ((1 << bits) - 1));
		cBitsWritten = cBitsWritten + bits;
		
		// push from buffer to file
		while (cBufferSize >= cBitsUsed) {
			int numberToWrite;
			
			numberToWrite = ((int) (cBaseByte + ((cBufferValue >> (cBufferSize - cBitsUsed)) & cValueMask)));
			
			cOutputArray.write(numberToWrite);
			
			// remove from buffer
			cBufferSize -= cBitsUsed;
			cBufferValue &= (1 << cBufferSize) - 1;
		}
	}
	
	/**
	Writes the boolean as a bit, 1 for true, 0 for false.
	*/
	public void pushBoolean(boolean value) throws IOException {
		if (value)
			pushInt(1, 1);
		else
			pushInt(1, 0);
	}
	
	/**
	Writes the string with 12 length bits (max length 4095 chars)
	bits means bits per character, should be 7-16.
	Fewer bits produces a smaller but less compatible file.
	*/
	public void pushString(int bits, String value) throws IOException {
		char[]						character;
		int							length;
		int							characterCounter;
		
		length = value.length();
		
		if (bits < 7 || bits > 16)
			throw new IOException("GEBitWriter.pushString: incorrect number of bits (" + bits + ") to push, min = 7, max = 16; bits written: " + getBitsWritten());
		if (length < 0)
			throw new IOException("GEBitWriter.pushString: incorrect length (" + length + ") of string to push; bits written: " + getBitsWritten());
		
		character = new char[length];
		
		pushInt(12, length);
		pushInt(5, bits);
		
		value.getChars(0, length, character, 0);
		
		for (characterCounter = 0; characterCounter < length; characterCounter++)
			pushInt(bits, ((int) character[characterCounter]));
	}
	
	/**
	Writes the string in 16-bit unicode format.
	*/
	public void pushString(String value) throws IOException {
		pushString(16, value);
	}
	
	/**
	Compresses an array of integers and stores it. It uses
	huffman compression and run length encoding, intended
	for compressing simple graphics and levels.
	<p>
	Only bitsUsed least significant bits of an element in
	data will be stored.
	*/
	public void compress(int[] data, int bitsUsed) throws IOException {
		if (data == null)
			throw new IOException("GEBitWriter.compress: no data to compress");
		
		// data converted to table indices
		int[]						dataIndex;
		
		// the unique values in data and the number of times they occur
		int[]						uniqueValue;
		int[]						uniqueValueFrequency;
		int							uniqueValueCount;
		
		int[]						nodeBits;
		int[]						nodeValue;
		int							mostBits;
		int[]						nodeWithBits;
		
		int							nodeCount;
		int							nodeCounter;
		
		{ // figure out all unique values and their frequency
			boolean					isUnique;
			int						valueCounter;
			int						uniqueValueCounter;
			
			uniqueValue = new int[data.length];
			uniqueValueFrequency = new int[data.length];
			
			// first element must be unique
			uniqueValue[0] = data[0];
			uniqueValueFrequency[0] = 1;
			uniqueValueCount = 1;
			
			// check all other elements
			for (valueCounter = 1; valueCounter < data.length; valueCounter++) {
				isUnique = true;
				
				for (uniqueValueCounter = 0; uniqueValueCounter < uniqueValueCount; uniqueValueCounter++)
					if (data[valueCounter] == uniqueValue[uniqueValueCounter]) {
						uniqueValueFrequency[uniqueValueCounter]++;
						isUnique = false;
					}
				
				if (isUnique) {
					uniqueValue[uniqueValueCount] = data[valueCounter];
					uniqueValueFrequency[uniqueValueCount]++;
					uniqueValueCount++;
				}
			}
		}
		
		if (uniqueValueCount == 1) {
			// only one unique value yields very simple compresion
			pushInt(bitsUsed, uniqueValueCount);
			pushInt(bitsUsed, uniqueValue[0]);
			return;
		}
		
		{ // trim arrays
			int[]					old;
			
			old = uniqueValue;
			uniqueValue = new int[uniqueValueCount];
			System.arraycopy(old, 0, uniqueValue, 0, uniqueValueCount);
			
			old = uniqueValueFrequency;
			uniqueValueFrequency = new int[uniqueValueCount];
			System.arraycopy(old, 0, uniqueValueFrequency, 0, uniqueValueCount);
		}
		
		// sort results by frequency
		sort(uniqueValueFrequency, uniqueValue);
		
		{	// create binary tree of nodes
			// the first uniqueValueCount nodes are the leaves i.e. unique values
			// the last node is the top of the hierarcy, not connected to
			int[]					nodeLeft;
			int[]					nodeRight;
			int[]					nodeFrequency;
			boolean[]				nodeUsed;
			int						smallestIndex;
			int						secondSmallestIndex;
			int						nodeCreatorCounter;
			
			nodeCount = uniqueValueCount * 2 - 1;
			
			// clear node data
			nodeLeft = new int[nodeCount];
			nodeRight = new int[nodeCount];
			nodeFrequency = new int[nodeCount];
			nodeUsed = new boolean[nodeCount];
			nodeBits = new int[nodeCount];
			nodeValue = new int[nodeCount];
			
			for (nodeCounter = 0; nodeCounter < uniqueValueCount; nodeCounter++) {
				nodeLeft[nodeCounter] = -1;
				nodeRight[nodeCounter] = -1;
				nodeFrequency[nodeCounter] = uniqueValueFrequency[nodeCounter];
				nodeUsed[nodeCounter] = false;
				nodeBits[nodeCounter] = -1;
			}
			for (nodeCounter = uniqueValueCount; nodeCounter < nodeCount; nodeCounter++) {
				nodeLeft[nodeCounter] = -1;
				nodeRight[nodeCounter] = -1;
				nodeFrequency[nodeCounter] = -1;
				nodeUsed[nodeCounter] = false;
				nodeBits[nodeCounter] = -1;
			}
			
			// figure the non-leaf nodes
			for (nodeCreatorCounter = uniqueValueCount; nodeCreatorCounter < nodeCount; nodeCreatorCounter++) {
				// find the two smallest available nodes
				smallestIndex = -1;
				secondSmallestIndex = -1;
				
				for (nodeCounter = 0; nodeCounter < nodeCreatorCounter; nodeCounter++) {
					if (!nodeUsed[nodeCounter]) {
						if (smallestIndex < 0)
							smallestIndex = nodeCounter;
						else if (secondSmallestIndex < 0)
							secondSmallestIndex = nodeCounter;
						else {
							// check new node against the larger of the two smallestIndices
							if (nodeFrequency[smallestIndex] > nodeFrequency[secondSmallestIndex]) {
								if (nodeFrequency[nodeCounter] < nodeFrequency[smallestIndex])
									smallestIndex = nodeCounter;
							} else {
								if (nodeFrequency[nodeCounter] < nodeFrequency[secondSmallestIndex])
									secondSmallestIndex = nodeCounter;
							}
						}
					}
				}
				
				// connect the two smallest nodes to this one
				nodeLeft[nodeCreatorCounter] = smallestIndex;
				nodeRight[nodeCreatorCounter] = secondSmallestIndex;
				nodeFrequency[nodeCreatorCounter] = nodeFrequency[nodeLeft[nodeCreatorCounter]] +
						nodeFrequency[nodeRight[nodeCreatorCounter]];
				nodeUsed[nodeLeft[nodeCreatorCounter]] = true;
				nodeUsed[nodeRight[nodeCreatorCounter]] = true;
			}
			
			// figure out number of bits for storing each node
			nodeBits[nodeCount - 1] = 0;
			for (nodeCounter = nodeCount - 1; nodeCounter >= 0; nodeCounter--) {
				if (nodeLeft[nodeCounter] >= 0)
					nodeBits[nodeLeft[nodeCounter]] = nodeBits[nodeCounter] + 1;
				if (nodeRight[nodeCounter] >= 0)
					nodeBits[nodeRight[nodeCounter]] = nodeBits[nodeCounter] + 1;
			}
		}
		
		{	// figure out compressed value for each leaf node
			// later nodes will have higher value
			int						currentNodeValue;
			
			mostBits = nodeBits[0];
			currentNodeValue = 0;
			
			for (nodeCounter = 0; nodeCounter < uniqueValueCount; nodeCounter++) {
				nodeValue[nodeCounter] = currentNodeValue;
				currentNodeValue += 1 << (mostBits - nodeBits[nodeCounter]);
			}
		}
		
		// build a table for restoring node values
		nodeWithBits = new int[mostBits]; // number of non-leaf nodes with index bits
		for (nodeCounter = 0; nodeCounter < mostBits; nodeCounter++)
			nodeWithBits[nodeCounter] = 0;
		
		// add up all non-leaf nodes and not top node which always exists
		
		for (nodeCounter = uniqueValueCount; nodeCounter < nodeCount - 1; nodeCounter++)
			nodeWithBits[nodeBits[nodeCounter]]++;
		
		{	// convert data to table indices
			int						valueCounter;
			int						uniqueValueCounter;
			
			dataIndex = new int[data.length];
			for (valueCounter = 0; valueCounter < data.length; valueCounter++)
				for (uniqueValueCounter = 0; uniqueValueCounter < uniqueValueCount; uniqueValueCounter++)
					if (data[valueCounter] == uniqueValue[uniqueValueCounter])
						dataIndex[valueCounter] = uniqueValueCounter;
		}
		
		{	// write to output file
			int						valueCounter;
			int						bitCounter;
			
			// write number of unique values
			pushInt(bitsUsed, uniqueValueCount);
			
			// write unique values
			for (valueCounter = 0; valueCounter < uniqueValueCount; valueCounter++)
				pushInt(bitsUsed, uniqueValue[valueCounter]);
			
			// right-shift node values
			for (nodeCounter = 0; nodeCounter < uniqueValueCount; nodeCounter++)
				nodeValue[nodeCounter] = nodeValue[nodeCounter] >> (mostBits - nodeBits[nodeCounter]);
			
			// write node table
			/*
			
			the table for restoring the nodes 000, 001, 01 and 1 would look like
			
			     / \
			    0   1
			   / \
			  0   1
			 / \
			0   1
			
			and stored with a 0 wherever there is a non-top, non-leaf node,
			and a 1 for every linebreak, i.e. 01011
			*/
			
			for (nodeCounter = 1; nodeCounter < mostBits; nodeCounter++) {
				for (bitCounter = 0; bitCounter < nodeWithBits[nodeCounter]; bitCounter++)
					pushBoolean(false);
				pushBoolean(true);
			}
			pushBoolean(true);
			
			// write data
			{
				int					savedValues;
				int					duplicates; // copies of dataIndex[savedValues]
				
				savedValues = 0;
				while (savedValues < data.length) {
					// figure out number of duplicates
					duplicates = 0;
					while (savedValues + duplicates + 1 < data.length &&
							duplicates < 32 &&
							dataIndex[savedValues] == dataIndex[savedValues + duplicates + 1])
						duplicates++;
					
					// compare bit lengths of RLE to flat
					if (6 + nodeBits[dataIndex[savedValues]] < (nodeBits[dataIndex[savedValues]] + 1) * (duplicates + 1)) { // 5 == header size
						// RLE
						pushBoolean(true);
						pushInt(5, duplicates - 1); // duplicates = 0 will never happen
						pushInt(nodeBits[dataIndex[savedValues]], nodeValue[dataIndex[savedValues]]);
					} else {
						// flat
						for (valueCounter = 0; valueCounter < duplicates + 1; valueCounter++) {
							pushBoolean(false);
							pushInt(nodeBits[dataIndex[savedValues]], nodeValue[dataIndex[savedValues]]);
						}
					}
					
					savedValues += duplicates + 1;
				}
			}
		}
	}
	
	/**
	Sorts a, and modifies the order of b to keep index pairs intact.
	*/
	protected void sort(int[] a, int[] b) {
		int n = a.length;
		int incr = n / 2;
		while (incr >= 1) {
			for (int i = incr; i < n; i++) {
				int tempa = a[i];
				int tempb = b[i];
				int j = i;
				while (j >= incr && tempa < a[j - incr]) {
					a[j] = a[j - incr];
					b[j] = b[j - incr];
					j -= incr;
				}
				a[j] = tempa;
				b[j] = tempb;
			}
			incr /= 2;
		}
	}
}

