import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;


/*
Handles a list of thumbnails/icons, provided by a GEImageCanvasOwner.
Supports horizontal or vertical display, with length and thickness
matching width and height when horizontal, and height and length when
vertical.
*/

public class GEImageCanvas extends JComponent implements AdjustmentListener, MouseListener
{
	public static final int			kAroundMargin = 4;
	public static final int			kBetweenMargin = 3;
	public static final int			kUnitIncrement = 16;
	public static final int			kMaxThickness = 150;
	
	
	protected boolean				cHorizontal;
	protected JScrollBar			cScrollBar;
	protected int					cPressedImage;
	protected int					cSelectedImage;
	protected int					cBufferLength;
	protected int					cBufferThickness;
	protected int[]					cBufferPixels;
	protected MemoryImageSource		cBufferSource;
	protected Image					cBuffer;
	protected GEImageCanvasOwner	cOwner;
	
	
	public GEImageCanvas(JScrollBar scrollBar, int bufferThickness, GEImageCanvasOwner owner) {
		cScrollBar = scrollBar;
		cOwner = owner;
		cPressedImage = -1;
		
		if (cScrollBar != null) {
			cScrollBar.addAdjustmentListener(this);
			cScrollBar.setUnitIncrement(kUnitIncrement);
			cHorizontal = (cScrollBar.getOrientation() == JScrollBar.HORIZONTAL);
		} else
			cHorizontal = true;
		
		addMouseListener(this);
		
		cSelectedImage = 0;
		
		setBufferThickness(bufferThickness);
	}
	
	public GEImageCanvas(JScrollBar scrollBar, GEImageCanvasOwner owner) {
		this(scrollBar, 16, owner);
	}
	
	public void paint(Graphics theGraphics) {
		Rectangle					bounds;
		int							imageOffset;
		int							imageCounter;
		Rectangle					selection = null;
		int							pixelCounter;
		int							factorX = 0;
		int							factorY = 0;
		
		bounds = getBounds();
		
		if (cHorizontal)
			factorX = 1;
		else
			factorY = 1;
		
		// erase background
		theGraphics.setColor(Color.white);
		theGraphics.fillRect(0, 0, bounds.width, bounds.height);
		theGraphics.setColor(Color.black);
		
		// erase buffer
		for (pixelCounter = 0; pixelCounter < cBufferPixels.length; pixelCounter++)
			cBufferPixels[pixelCounter] = 0xFFFFFFFF;
		
		// draw images
		if (cScrollBar != null)
			imageOffset = -cScrollBar.getValue() + kAroundMargin;
		else
			imageOffset = kAroundMargin;
		
		if (getImageCount() > 0) {
			for (imageCounter = 0; imageCounter < getImageCount(); imageCounter++) {
				if (imageOffset + getImageLength(imageCounter) >= 0 &&
						imageOffset < bounds.width * factorX + bounds.height * factorY)
					getImage(imageCounter).draw(cBufferPixels, getBufferWidth(), imageOffset * factorX, imageOffset * factorY, 0);
				
				// define selection if appropriate
				if (imageCounter == getSelected())
					selection = new Rectangle(imageOffset * factorX, imageOffset * factorY, getImageWidth(imageCounter) + 1, getImageHeight(imageCounter) + 1);
				
				imageOffset += getImageLength(imageCounter) + kBetweenMargin;
			}
			cBufferSource.newPixels();
		}
		
		// copy buffer to window
		theGraphics.drawImage(cBuffer, kAroundMargin * factorY, kAroundMargin * factorX, null);
		
		if (selection != null) { // draw selection
			theGraphics.drawRect(
					selection.x + kAroundMargin * factorY - 1,
					selection.y + kAroundMargin * factorX - 1,
					selection.width, selection.height);
		}
		
		// draw frame
		theGraphics.setColor(this.getBackground().darker());
		theGraphics.drawLine(0, 0, bounds.width - 1, 0);
		theGraphics.drawLine(0, 0, 0, bounds.height - 1);
		theGraphics.setColor(this.getBackground().brighter());
		theGraphics.drawLine(bounds.width - 1, 0, bounds.width - 1, bounds.height - 1);
		theGraphics.drawLine(0, bounds.height - 1, bounds.width - 1, bounds.height - 1);
		
		theGraphics.setColor(Color.black);
		theGraphics.drawRect(1, 1, bounds.width - 3, bounds.height - 3);
	}
	
	// Called when the contents may have changed size
	public void updateScrollBar()
	{
		int							imageCounter;
		int							totalLength;
		int							newValue;
		int							visibleLength;
		
		if (cScrollBar == null)
			return;
		
		if (cHorizontal)
			visibleLength = getBounds().width;
		else
			visibleLength = getBounds().height;
		
		totalLength = 2 * kAroundMargin - kBetweenMargin;
		for (imageCounter = 0; imageCounter < getImageCount(); imageCounter++)
			totalLength += getImageLength(imageCounter) + kBetweenMargin;
		
		newValue = cScrollBar.getValue();
		
		if (newValue + visibleLength > totalLength)
			newValue = totalLength - visibleLength;
		if (newValue < 0)
			newValue = 0;
		
		cScrollBar.setValues(newValue, visibleLength, 0, totalLength);
		cScrollBar.setBlockIncrement(visibleLength - kUnitIncrement);
		
		repaint();
	}
	
	public void setBufferThickness(int thickness) {
		if (thickness < 8)
			thickness = 8;
		if (thickness > kMaxThickness)
			thickness = kMaxThickness;
		
		thickness += 2 * kAroundMargin;
		
		if (thickness == cBufferThickness)
			return;
		
		cBufferThickness = thickness;
		
		if (cHorizontal) {
			setPreferredSize(new Dimension(2000, cBufferThickness));
			setMinimumSize(new Dimension(50, cBufferThickness));
			setMaximumSize(new Dimension(2000, cBufferThickness));
		} else {
			setPreferredSize(new Dimension(cBufferThickness, 2000));
			setMinimumSize(new Dimension(cBufferThickness, 50));
			setMaximumSize(new Dimension(cBufferThickness, 2000));
		}
		
		revalidate();
	}
	
	// Interface with owner, convenience methods
	public int getImageCount() {
		return cOwner.getImageCount(this);
	}
	
	public GPixels getImage(int imageIndex) {
		return cOwner.getImage(this, imageIndex);
	}
	
	public void newSelectedImage() {
		cOwner.newSelectedImage(this);
	}
	
	public void clickOnSelectedImage() {
		cOwner.clickOnSelectedImage(this);
	}
	
	public void draggedImage(int fromIndex, int toIndex) {
		cOwner.draggedImage(this, fromIndex, toIndex);
	}
	
	public int getImageWidth(int imageIndex) {
		return getImage(imageIndex).getWidth();
	}
	
	public int getImageHeight(int imageIndex) {
		return getImage(imageIndex).getHeight();
	}
	
	public int getImageLength(int imageIndex) {
		if (cHorizontal)
			return getImage(imageIndex).getWidth();
		else
			return getImage(imageIndex).getHeight();
	}
	
	public int getImageThickness(int imageIndex) {
		if (cHorizontal)
			return getImage(imageIndex).getHeight();
		else
			return getImage(imageIndex).getWidth();
	}
	
	public void setSelected(int newSelected) {
		cSelectedImage = newSelected;
		if (cSelectedImage >= getImageCount())
			cSelectedImage = getImageCount() - 1;
		if (cSelectedImage < 0)
			cSelectedImage = 0;
		
		scrollToSelected();
		repaint();
	}
	
	// automatically scrolls to display the selected image
	public void scrollToSelected() {
		int							imageOffset;
		int							imageCounter;
		int							selectionMin;
		int							selectionMax;
		
		if (cScrollBar == null)
			return;
		
		selectionMin = 0;
		selectionMax = 0;
		
		imageOffset = kAroundMargin;
		for (imageCounter = 0; imageCounter < getImageCount(); imageCounter++) {
			if (imageCounter == getSelected()) {
				selectionMin = imageOffset - kAroundMargin;
				selectionMax = imageOffset + getImageLength(imageCounter) + kAroundMargin;
				break;
			}
			
			imageOffset += getImageLength(imageCounter) + kBetweenMargin;
		}
		
		if (selectionMax > (cScrollBar.getValue() + cScrollBar.getVisibleAmount())) {
			cScrollBar.setValue(selectionMax - cScrollBar.getVisibleAmount());
			repaint();
		}
		if (selectionMin < cScrollBar.getValue()) {
			cScrollBar.setValue(selectionMin);
			repaint();
		}
	}
	
	public int getBufferWidth() {
		if (cHorizontal)
			return cBufferLength;
		else
			return cBufferThickness;
	}
	
	public int getBufferHeight() {
		if (cHorizontal)
			return cBufferThickness;
		else
			return cBufferLength;
	}
	
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		
		if (cHorizontal) {
			cBufferLength = width;
			cBufferThickness = height;
		} else {
			cBufferLength = height;
			cBufferThickness = width;
		}
		
		if (cBufferLength <= 16)
			cBufferLength = 16;
		if (cBufferThickness <= 16)
			cBufferThickness = 16;
		
		cBufferPixels = new int[cBufferLength * cBufferThickness];
		cBufferSource = new MemoryImageSource(getBufferWidth(), getBufferHeight(), cBufferPixels, 0, getBufferWidth());
		cBufferSource.setAnimated(true);
		cBuffer = createImage(cBufferSource);
		updateScrollBar();
		scrollToSelected();
	}
	
	public int getSelected() {
		if (cSelectedImage >= getImageCount())
			cSelectedImage = getImageCount() - 1;
		if (cSelectedImage < 0)
			cSelectedImage = 0;
		
		return cSelectedImage;
	}
	
	public void adjustmentValueChanged(AdjustmentEvent theEvent) {
		repaint();
	}
	
	public int getImageIndex(int offset) {
		// accepts mouse length coordinate
		int								localOffset;
		int								imageCounter;
		int								imageOffset;
		int								length;
		
		if (getImageCount() <= 0)
			return -1;
		
		if (cHorizontal)
			length = getBounds().width;
		else
			length = getBounds().height;
		
		if (offset < 0)
			return -1;
		if (offset > length)
			return getImageCount();
		
		// find which image was clicked
		localOffset = (offset - kAroundMargin + cScrollBar.getValue());
		imageOffset = 0;
		imageCounter = 0;
		
		while (imageCounter < getImageCount()) {
			imageOffset += getImageLength(imageCounter) + kBetweenMargin;
			
			if (imageOffset > localOffset)
				break;
			
			imageCounter++;
		}
		
		if (imageCounter >= getImageCount()) // past last image
			return getImageCount();
		
		return imageCounter;
	}
	
	public void mousePressed(MouseEvent theEvent) {
		if (cHorizontal)
			cPressedImage = getImageIndex(theEvent.getX());
		else
			cPressedImage = getImageIndex(theEvent.getY());
		repaint();
	}
	
	public void mouseReleased(MouseEvent theEvent) {
		int								imageIndex;
		
		if (cHorizontal)
			imageIndex = getImageIndex(theEvent.getX());
		else
			imageIndex = getImageIndex(theEvent.getY());
		
		if (cPressedImage == -1 || cPressedImage >= getImageCount()) {
			// invalid pressed index
		} else if (imageIndex == cPressedImage) {
			// pressed and released on the same image
			if (imageIndex == cSelectedImage) {
				clickOnSelectedImage();
			} else {
				cSelectedImage = imageIndex;
				newSelectedImage();
			}
		} else {
			// dragged from one index to another
			if (theEvent.isAltDown())
				draggedImage(cPressedImage, imageIndex);
		}
		
		cPressedImage = -1;
		repaint();
	}
	
	public void mouseClicked(MouseEvent theEvent) {}
	public void mouseEntered(MouseEvent theEvent) {}
	public void mouseExited(MouseEvent theEvent) {}
}

