import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;


/*
Handles editing of levels, contains a tile and sprite palette,
and a GELevelCanvas that displays a level.

GELevelFrame will release buffer memory when closing, and
reallocate it when opening, becuase the buffer can take several
megabytes per level.

*/


public class GELevelFrame extends JFrame implements ActionListener, GEImageCanvasOwner
{
	protected final boolean			kVerticalPalette = false;
	
	protected GEImageCanvas			cTileSpriteCanvas; // palette
	protected JPanel				cLevelDetailPanel;
	protected GELevelCanvas			cLevelDetailCanvas;
	protected JScrollBar			cLevelHorizontalScrollBar;
	protected JScrollBar			cLevelVerticalScrollBar;
	protected GLevel				cLevel;
	
	protected boolean				cBufferDisabled;
	
	
	public GELevelFrame(GLevel level, String title) {
		JPanel						contents;
		JPanel						tileSpritePanel;
		JScrollBar					tileSpriteScrollBar;
		GridBagLayout				gridBag;
		GridBagConstraints			constraints;
		JMenuBar					menuBar;
		
		setTitle(title);
		setLocation(30, 30);
		setSize(300, 350);
		
		cLevel = level;
		
		// menu bar
		menuBar = new JMenuBar();
		setJMenuBar(menuBar);
		
		menuBar.add(GEditor.createMenu("Level", new Object[] {"Add Margins", "Show Information"},
				new int[] {KeyEvent.VK_M, KeyEvent.VK_I}, this));
		
		// panels
		contents = new JPanel(new BorderLayout(3, 3));
		contents.setBorder(new EmptyBorder(3, 3, 3, 3));
		
		// tile and sprite palette
		tileSpritePanel = new JPanel(new BorderLayout());
		
		if (kVerticalPalette) {
			tileSpriteScrollBar = new JScrollBar(JScrollBar.VERTICAL);
			tileSpritePanel.add(tileSpriteScrollBar, BorderLayout.EAST);
			cTileSpriteCanvas = new GEImageCanvas(tileSpriteScrollBar, GTileForm.getTileWidth(), this);
			tileSpritePanel.add(cTileSpriteCanvas, BorderLayout.WEST);
			contents.add(tileSpritePanel, BorderLayout.WEST);
		} else {
			tileSpriteScrollBar = new JScrollBar(JScrollBar.HORIZONTAL);
			tileSpritePanel.add(tileSpriteScrollBar, BorderLayout.SOUTH);
			cTileSpriteCanvas = new GEImageCanvas(tileSpriteScrollBar, GTileForm.getTileHeight(), this);
			tileSpritePanel.add(cTileSpriteCanvas, BorderLayout.NORTH);
			contents.add(tileSpritePanel, BorderLayout.NORTH);
		}
		
		// level display
		gridBag = new GridBagLayout();
		constraints = new GridBagConstraints();
		
		cLevelDetailPanel = new JPanel(gridBag);
		
		cLevelHorizontalScrollBar = new JScrollBar(JScrollBar.HORIZONTAL);
		cLevelVerticalScrollBar = new JScrollBar(JScrollBar.VERTICAL);
		cLevelDetailCanvas = new GELevelCanvas(cLevel, cTileSpriteCanvas, cLevelHorizontalScrollBar, cLevelVerticalScrollBar, this);
		cBufferDisabled = false;
		
		// levelcanvas
		constraints.gridx = 0;
		constraints.gridy = 0;
		constraints.gridwidth = 1;
		constraints.gridheight = 1;
		constraints.weightx = 100;
		constraints.weighty = 100;
		constraints.fill = GridBagConstraints.BOTH;
		
		cLevelDetailPanel.add(cLevelDetailCanvas, constraints);
		
		// horizontal scroll bar
		constraints.gridx = 0;
		constraints.gridy = 1;
		constraints.weightx = 0;
		constraints.weighty = 0;
		
		cLevelDetailPanel.add(cLevelHorizontalScrollBar, constraints);
		
		// vertical scroll bar
		constraints.gridx = 1;
		constraints.gridy = 0;
		cLevelDetailPanel.add(cLevelVerticalScrollBar, constraints);
		
		contents.add(cLevelDetailPanel, BorderLayout.CENTER);
		
		getContentPane().add(contents);
	}
	
	public void show() {
		super.show();
		// reallocate buffer memory
		cLevelDetailCanvas.refresh();
		cBufferDisabled = false;
	}
	
	public void setVisible(boolean b) {
		if (b) {
			super.setVisible(b);
			// reallocate buffer memory
			cLevelDetailCanvas.refresh();
			cBufferDisabled = false;
		} else {
			((GELevel) cLevel).createThumbnail();
			GEditor.sInstance.updateImageCanvas(false);
			((GELevel) cLevel).frameDisposed();
			// release buffer memory
			cLevelDetailCanvas.getLevelCanvas().setSize(0, 0);
			cBufferDisabled = true;
			super.setVisible(b);
		}
	}
	
	public void updateImageCanvas(boolean resize) {
		if (resize)
			cTileSpriteCanvas.setBufferThickness(GTileForm.getTileHeight());
		cTileSpriteCanvas.updateScrollBar();
		cTileSpriteCanvas.repaint();
		
		if (isShowing())
			cLevelDetailCanvas.refresh();
	}
	
	public void refresh() {
		cTileSpriteCanvas.updateScrollBar();
		if (isShowing())
			cLevelDetailCanvas.refresh();
	}
	
	public int getImageCount(GEImageCanvas canvas) {
		return GTileForm.getCount() + GSpriteForm.getCount();
	}
	
	public GPixels getImage(GEImageCanvas canvas, int imageIndex) {
		if (imageIndex < GTileForm.getCount())
			return GTileForm.get(imageIndex).getImage(0);
		else
			return GSpriteForm.get(imageIndex - GTileForm.getCount()).getImage(0);
	}
	
	public void newSelectedImage(GEImageCanvas canvas) {}
	
	public void clickOnSelectedImage(GEImageCanvas canvas) {}
	
	public void draggedImage(GEImageCanvas canvas, int fromIndex, int toIndex) {
		// dragging in tiles/sprites not supported
		canvas.setSelected(toIndex);
	}
	
	public void addMargins() {
		Object[]					answer;
		int							left;
		int							top;
		int							right;
		int							bottom;
		
		answer = GEAskDialog.ask(this, 200, "Add Margins",
			new String[] {"Left", "Top", "Right", "Bottom"},
			new Object[] {new Integer(0),
				new Integer(0),
				new Integer(0),
				new Integer(0)});
		
		if (answer != null) {
			left = ((Integer) answer[0]).intValue();
			top = ((Integer) answer[1]).intValue();
			right = ((Integer) answer[2]).intValue();
			bottom = ((Integer) answer[3]).intValue();
			
			((GELevel) cLevel).addMargins(left, top, right, bottom);
			cLevelDetailCanvas.refresh();
			
			((GELevel) cLevel).createThumbnail();
			GEditor.sInstance.updateImageCanvas(true);
		}
	}
	
	public void showInformation() {
		GEAskDialog.ask(this, 300, "Information (not editable)",
			new String[] {
				"Number of sprites", 
				"Width (tiles)", 
				"Height (tiles)", 
				"Width (pixels)", 
				"Height (pixels)"},
			new String[] {
				"" + cLevel.getSpriteCount(), 
				"" + cLevel.getWidth(), 
				"" + cLevel.getHeight(), 
				"" + cLevel.getWidth() * GTileForm.getTileWidth(), 
				"" + cLevel.getHeight() * GTileForm.getTileHeight()});
	}
	
	public void actionPerformed(ActionEvent theEvent) {
		String						command;
		
		command = theEvent.getActionCommand();
		
		if (command.equals("Add Margins"))
			addMargins();
		else if (command.equals("Show Information"))
			showInformation();
	}
}

