package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.net.*;
import java.io.*;


/**
 * GBitReader is used to conveniently read unsigned integers
 * with different byte sizes, booleans, strings, and compressed
 * arrays of integers.
 * <p>
 * GBitReader uses the 6 least significant bits (bytes in the
 * range 0x40 - 0x7F), because of problems with cross-platform
 * compatibility using the bytes 0x80 - 0xFF and line-break
 * characters (0x0D). Specifically, some web browsers tend to
 * convert them from one character set to another.
 * <p>
 * Despite the unfortunate naming, GBitReader should in Java
 * terminology be considered an InputStream, not a Reader. (You
 * will not need to know the diffrence to use GBitReader.)
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GBitReader
{
	public static final int			kMaxIntBits = 32;
	
	protected int					cBitsUsed;
	protected int					cValueMask;
	
	protected DataInputStream		cInput;
	protected int					cBufferValue; // least significant bits are used
	protected short					cBufferSize; // size in bits
	protected int					cBitsRead;
	
	/**
	 * Opens a file specified by a File.
	 */
	public GBitReader(File data) throws IOException {
		cInput = new DataInputStream(new BufferedInputStream(new FileInputStream(data)));
		
		cBufferValue = 0;
		cBufferSize = 0;
		cBitsRead = 0;
		
		// uses only 6 least significant bits
		cBitsUsed = 6;
		cValueMask = 0x3F;
	}
	
	/**
	 * Opens a file specified by an URL.
	 * Warning: The constructor can block if openStream() runs into problems.
	 */
	public GBitReader(URL fileLocation) throws IOException {
		cInput = new DataInputStream(new BufferedInputStream(fileLocation.openStream()));
		
		cBufferValue = 0;
		cBufferSize = 0;
		cBitsRead = 0;
		
		cBitsUsed = 6;
		cValueMask = 0x3F;
	}
	
	/**
	 * Opens a file specified by a file name.
	 */
	public GBitReader(String fileName) throws IOException {
		this(new File(fileName));
	}
	
	public void close() throws IOException {
		if (cInput == null)
			throw new IOException("GBitReader.close(): no valid file to close");
		
		cInput.close();
		cInput = null;
	}
	
	/**
	 * Returns the number of bits read from the file. Only six bits
	 * are read per byte.
	 */
	public int getBitsRead() {
		return cBitsRead;
	}
	
	/**
	 * Reads a specified number of bits and returns the unsigned
	 * integer they represent.
	 */
	public int pullInt(int bits) throws IOException {
		int							returnValue;
		
		if (cInput == null)
			throw new IOException("GBitReader.pullInt(): no valid file to read from");
		
		if (bits < 1 || bits > kMaxIntBits)
			throw new IOException("GBitReader.pullInt(): incorrect number of bits (" + bits + ") to read, min = 1, max = " + kMaxIntBits + "; bits read " + getBitsRead());
		
		if (bits > 16) {
			returnValue = pullInt(bits - 16) << 16;
			bits = 16;
		} else
			returnValue = 0;
		
		cBitsRead += bits;
		
		// pull from data to buffer
		while (cBufferSize < bits)
		{
			cBufferValue <<= cBitsUsed;
			cBufferSize += cBitsUsed;
			
			cBufferValue += cInput.readByte() & cValueMask;
		}
		
		// prepare return value
		returnValue += (cBufferValue >> (cBufferSize - bits)) & ((1 << bits) - 1);
		
		// remove return value from buffer
		cBufferSize -= bits;
		cBufferValue &= (1 << cBufferSize) - 1;
		
		return returnValue;
	}
	
	/**
	 * Reads one bit from file and returns true for 1, false for 0.
	 */
	public boolean pullBoolean() throws IOException {
		return (pullInt(1) != 0);
	}
	
	/**
	 * Reads a string stored with GEBitWriter.pushString().
	 */
	public String pullString() throws IOException {
		char[]						character;
		int							length;
		int							bits;
		int							characterCounter;
		
		length = pullInt(12);
		bits = pullInt(5);
		
		if (bits < 1 || bits > 16)
			throw new IOException("GBitReader.pullString(): incorrect number of bits per character (" + bits + "); bits read: " + getBitsRead());
		
		character = new char[length];
		
		for (characterCounter = 0; characterCounter < length; characterCounter++)
			character[characterCounter] = (char) pullInt(bits);
		
		return new String(character);
	}
	
	/**
	 * Reads, decompresses and returns an array of integers.
	 * bitsUsed must match the bitsUsed when compressed.
	 */
	public int[] decompress(int valueCount, int bitsUsed) throws IOException {
		int							uniqueValueCount;
		int[]						data;
		int[]						uniqueValue;
		int[]						nodeWithBits;
		int[]						nodeBits;
		int[]						nodeValue;
		int							mostBits;
		int							bitCounter;
		int							remainingSlots;
		int							currentNode;
		int							openSlots;
		int							nextNodeValue;
		int							valueCounter;
		int							nodeCounter;
		int							readElementValue;
		int							readElementBitCount;
		boolean						foundMatch;
		
		// read from input file
		uniqueValueCount = pullInt(bitsUsed);
		
		if (uniqueValueCount == 1) {
			// only one value, all elements will have that value
			int						singleUniqueValue;
			
			singleUniqueValue = pullInt(bitsUsed);
			
			data = new int[valueCount];
			for (valueCounter = 0; valueCounter < valueCount; valueCounter++)
				data[valueCounter] = singleUniqueValue;
			
			return data;
		}
		
		// read unique values
		uniqueValue = new int[uniqueValueCount];
		for (valueCounter = 0; valueCounter < uniqueValueCount; valueCounter++)
			uniqueValue[valueCounter] = pullInt(bitsUsed);
		
		// read table of nodes
		nodeWithBits = new int[uniqueValueCount + 1]; // uVC + 1 should always be large enough
		for (bitCounter = 0; bitCounter < nodeWithBits.length; bitCounter++)
			nodeWithBits[bitCounter] = 0;
		
		nodeValue = new int[uniqueValueCount];
		nodeBits = new int[uniqueValueCount];
		
		mostBits = 0;
		bitCounter = 0;
		remainingSlots = 1;
		while (mostBits == 0) {
			bitCounter++;
			remainingSlots *= 2;
			openSlots = 0;
			
			while (pullBoolean() == false)
				openSlots++;
			
			nodeWithBits[bitCounter] = remainingSlots - openSlots;
			
			remainingSlots = openSlots;
			
			if (remainingSlots == 0)
				mostBits = bitCounter;
		}
		
		// figure out the compressed values of the nodes
		nextNodeValue = 0;
		nodeCounter = 0;
		for (bitCounter = mostBits; bitCounter > 0; bitCounter--) {
			if (nodeWithBits[bitCounter] != 0) {
				for (valueCounter = 0; valueCounter < nodeWithBits[bitCounter]; valueCounter++) {
					nodeValue[nodeCounter] = nextNodeValue;
					nodeBits[nodeCounter] = bitCounter;
					nodeCounter++;
					nextNodeValue += 1 << (mostBits - bitCounter);
				}
			}
		}
		
		// right-shift node values
		for (nodeCounter = 0; nodeCounter < uniqueValueCount; nodeCounter++)
			nodeValue[nodeCounter] = nodeValue[nodeCounter] >> (mostBits - nodeBits[nodeCounter]);
		
		// read the data
		data = new int[valueCount];
		for (valueCounter = 0; valueCounter < valueCount; valueCounter++)
			data[valueCounter] = -1;
		
		{
			int						readValues;
			int						duplicates;
			boolean					RLE;
			int						dataValue;
			
			readValues = 0;
			dataValue = 0;
			while (readValues < valueCount) {
				// determine compression version
				RLE = pullBoolean();
				if (RLE)
					duplicates = pullInt(5) + 1;
				else
					duplicates = 0;
				
				// read data value
				readElementValue = 0;
				readElementBitCount = 0;
				foundMatch = false;
				while (!foundMatch) {
					readElementValue <<= 1;
					readElementValue += pullInt(1);
					readElementBitCount++;
					for (nodeCounter = 0; nodeCounter < uniqueValueCount; nodeCounter++) {
						if (nodeBits[nodeCounter] == readElementBitCount &&
								nodeValue[nodeCounter] == readElementValue) {
							dataValue = uniqueValue[nodeCounter];
							foundMatch = true;
						}
					}
					if (readElementBitCount > mostBits)
						throw new IOException("GBitReader.decompress(): Failed to find match " + readValues);
				}
				
				for (valueCounter = 0; valueCounter < duplicates + 1; valueCounter++)
					data[readValues + valueCounter] = dataValue;
				
				readValues += duplicates + 1;
			}
		}
		
		return data;
	}
}

