package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.IOException;


/**
 * GFactory controls the creation of all GLevels, GPixels,
 * GProperties, GSprites, GSpriteForms and GTileForms.
 * If you want to subclass any of them, override the appropriate
 * method in GFactory to produce the subclass. Especially
 * GSprite is subclassed in many games, to provide custom sprite
 * behavior. Different GSprite subclasses can be produced
 * depending on the value of formIndex.
 * <p>
 * When appropriate, use the static methods in GFactory
 * (GFactory.createX()) to create your objects. Before initializing
 * any other part of GameForm (including loading a data file), use
 * GFactory.setFactory(new myFactory()) to implement your own
 * factory. All subsequent supported objects will be created by the
 * new factory.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GFactory
{
	protected static GFactory		sFactory;
	
	
	/**
	 * Override to produce custom GLevel objects.
	 */
	public GLevel newLevel() {
		return new GLevel();
	}
	
	/**
	 * Override to produce custom GPixels objects.
	 */
	public GPixels newPixels() {
		return new GPixels();
	}
	
	/**
	 * Override to produce custom GSprite objects. Different GSprite
	 * subclasses can be produced depending on the value of
	 * formIndex.
	 */
	public GSprite newSprite(int formIndex) {
		return new GSprite(formIndex);
	}
	
	/**
	 * Override to produce custom GSpriteForm objects.
	 */
	public GSpriteForm newSpriteForm() {
		return new GSpriteForm();
	}
	
	/**
	 * Override to produce custom GTileForm objects.
	 */
	public GTileForm newTileForm() {
		return new GTileForm();
	}
	
	/**
	 * Override to produce custom GProperties objects.
	 */
	public GProperties newProperties() {
		return new GProperties();
	}
	
	/**
	 * Makes factory the default factory, responsible for creating
	 * all the new supported objects.
	 */
	public static void setFactory(GFactory factory) {
		sFactory = factory;
	}
	
	public static GFactory getFactory() {
		if (sFactory == null)
			sFactory = new GFactory();
		
		return sFactory;
	}
	
	// GLevel factory, no need to override
	public static GLevel createLevel(GBitReader input) throws IOException {
		GLevel						level;
		
		level = createLevel();
		level.readFromFile(input);
		return level;
	}
	
	public static GLevel createLevel(GLevel form) {
		return (GLevel) form.clone();
	}
	
	public static GLevel createLevel() {
		return getFactory().newLevel();
	}
	
	// GPixels factory, no need to override
	public static GPixels createPixels(GBitReader input) throws IOException {
		GPixels						pixels;
		
		pixels = createPixels();
		pixels.readFromFile(input);
		return pixels;
	}
	
	public static GPixels createPixels(GPixels form) {
		return (GPixels) form.clone();
	}
	
	public static GPixels createPixels() {
		return getFactory().newPixels();
	}
	
	// GSprite factory, no need to override
	public static GSprite createSprite(int formIndex, GBitReader input) throws IOException {
		GSprite						sprite;
		
		sprite = createSprite(formIndex);
		sprite.readFromFile(input);
		return sprite;
	}
	
	public static GSprite createSprite(GSprite form) {
		return (GSprite) form.clone();
	}
	
	public static GSprite createSprite(int formIndex) {
		return getFactory().newSprite(formIndex);
	}
	
	// GSpriteForm factory, no need to override
	public static GSpriteForm createSpriteForm(GBitReader input) throws IOException {
		GSpriteForm					spriteForm;
		
		spriteForm = createSpriteForm();
		spriteForm.readFromFile(input);
		return spriteForm;
	}
	
	public static GSpriteForm createSpriteForm(GSpriteForm form) {
		return (GSpriteForm) form.clone();
	}
	
	public static GSpriteForm createSpriteForm() {
		return getFactory().newSpriteForm();
	}
	
	// GTileForm factory, no need to override
	public static GTileForm createTileForm(GBitReader input) throws IOException {
		GTileForm					tileForm;
		
		tileForm = createTileForm();
		tileForm.readFromFile(input);
		return tileForm;
	}
	
	public static GTileForm createTileForm(GTileForm form) {
		return (GTileForm) form.clone();
	}
	
	public static GTileForm createTileForm() {
		return getFactory().newTileForm();
	}
	
	// GProperties factory, no need to override
	public static GProperties createProperties(GBitReader input) throws IOException {
		GProperties					properties;
		
		properties = createProperties();
		properties.readFromFile(input);
		return properties;
	}
	
	public static GProperties createProperties(GProperties form) {
		return (GProperties) form.clone();
	}
	
	public static GProperties createProperties() {
		return getFactory().newProperties();
	}
}

