package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.net.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;


/**
 * GFileManager handles loading of data files. Invoking load()
 * will create two new threads, one will read from the specified
 * file and the other will report to a GLoadListener object, to
 * keep the user informed about the progress.
 * <p>
 * The specified GLoadListener's newProgress() method will
 * repeatedly be invoked during loading. It will also be invoked
 * with bitsRead equaling kFailed or kFinished when appropriate.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GFileManager implements Runnable {
	public static final int			kFailed = -2;
	public static final int			kFinished = -1;
	
	protected Thread				cReadThread;
	protected Thread				cReportThread;
	
	protected String				cMessage;
	protected int					cBitsRead;
	protected boolean				cFailed;
	protected boolean				cCompleted;
	
	protected GLoadListener			cListener;
	protected File					cInputFile;
	protected URL					cInputURL;
	protected GBitReader			cInput;
	
	
	/**
	 * Opens and reads a file specified by an URL.
	 * <p>
	 * If no listener is passed, this method will block and return an
	 * error code, or null if the file was read successfully.
	 * <p>
	 * If a listener is passed, this method returns null immediately,
	 * but the file is not read until listener.newProgress(kFinished)
	 * is invoked.
	 */
	public static String load(URL inputURL, GLoadListener listener) {
		return load(null, inputURL, listener);
	}
	
	/**
	 * Opens and reads a file specified by a File.
	 * <p>
	 * If no listener is passed, this method will block and return an
	 * error code, or null if the file was read successfully.
	 * <p>
	 * If a listener is passed, this method returns null immediately,
	 * but the file is not read until listener.newProgress(kFinished)
	 * is invoked.
	 */
	public static String load(File inputFile, GLoadListener listener) {
		return load(inputFile, null, listener);
	}
	
	/**
	 * Opens and reads a file specified by a file name.
	 * <p>
	 * If no listener is passed, this method will block and return an
	 * error code, or null if the file was read successfully.
	 * <p>
	 * If a listener is passed, this method returns null immediately,
	 * but the file is not read until listener.newProgress(kFinished)
	 * is invoked.
	 */
	public static String load(String inputFileName, GLoadListener listener) {
		return load(new File(inputFileName), null, listener);
	}
	
	protected static String load(File inputFile, URL inputURL, GLoadListener listener) {
		GFileManager				loader;
		
		loader = new GFileManager(inputFile, inputURL, listener);
		
		if (listener == null)
			while (!loader.cFailed && !loader.cCompleted) {
				try {
					Thread.sleep(200);
				} catch (InterruptedException theException) {
					return "Interrupted";
				}
			}
		
		return loader.cMessage;
	}
	
	private GFileManager(File inputFile, URL inputURL, GLoadListener listener) {
		cReadThread = null;
		cReportThread = null;
		
		cMessage = null;
		cBitsRead = 0;
		cFailed = false;
		cCompleted = false;
		
		cListener = listener;
		cInputFile = inputFile;
		cInputURL = inputURL;
		cInput = null;
		
		cReportThread = new Thread(this);
		cReportThread.setPriority(Thread.MIN_PRIORITY + 1);
		cReportThread.start();
	}
	
	public void run() {
		if (cReadThread == null) {
			// report thread
			int					lastBitsRead;
			int					deadTime;
			
			/* Start reading. GFileManager uses a slightly
			awkward trick to have two threads running on
			the same Runnable object. The first thread will
			be the reporting thread (currently running), and
			the second thread will be the reading thread
			(started immediately below). */
			cReadThread = new Thread(this);
			cReadThread.setPriority(Thread.MIN_PRIORITY);
			cReadThread.start();
			
			deadTime = 0;
			lastBitsRead = 0;
			
			while (true) {
				if (cCompleted || cFailed) {
					// reporting was completed when either flag was set
					return;
				} else if (cInput == null) {
					// reading not yet started
				} else if (lastBitsRead != cInput.getBitsRead()) {
					lastBitsRead = cInput.getBitsRead();
					deadTime = 0;
					
					if (cListener != null)
						cListener.newProgress(lastBitsRead);
				} else {
					deadTime++;
					if (deadTime > 60) { // six seconds
						if (cListener != null)
							cListener.newProgress(kFailed);
						
						cMessage = "Process timed out. (6s)";
						cFailed = true;
						return;
					}
				}
				
				try {
					Thread.sleep(100);
				} catch (InterruptedException theException) {
					if (cListener != null)
						cListener.newProgress(kFailed);
					
					cMessage = "Process timed out. (interrupted)";
					cFailed = true;
					return;
				}
			}
		} else {
			// read thread
			try {
				if (cInputURL != null)
					cInput = new GBitReader(cInputURL);
				else if (cInputFile != null)
					cInput = new GBitReader(cInputFile);
				else
					throw new IOException("GFileManager.readFromFile: Invalid input");
				
				readFromFile(cInput);
				
				cInput.close();
				
				if (cListener != null)
					cListener.newProgress(kFinished);
				cCompleted = true;
			} catch (IOException theException) {
				if (cInput != null)
					try { cInput.close(); } catch (IOException e) {}
				if (cListener != null)
					cListener.newProgress(kFailed);
				cMessage = "readFromFile() failed; " + theException;
				cFailed = true;
			}
		}
	}
	
	protected void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		GOutput.logln("reading...");
		
		if (input == null)
			throw new IOException("GFileManager.readFromFile: invalid file");
		
		version = input.pullInt(8);
		
		GOutput.logln("version " + version);
		
		if (version == 1) {
			GTileForm.readAllFromFile(input);
			GSpriteForm.readAllFromFile(input);
			GLevel.readAllFromFile(input);
			
			GOutput.logln("reading complete");
		} else {
			throw new IOException("GFileManager.readFromFile: Unknown version: " + version);
		}
	}
}


