package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GLevel is GameForm's level format. It stores the tiles and
 * sprites that make up the level. The tiles are stored as an array
 * of GTileForm indices, and the sprites are stored as an array
 * of GSprite objects.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GLevel implements Cloneable
{
	protected static GLevel[]		sLevel;
	
	public static final int			kMaxWidth = 4095;
	public static final int			kMaxHeight = 4095;
	
	protected static int			sNumberCount;
	protected static int			sFlagCount;
	protected static int			sTextCount;
	
	protected boolean				cDisableAddSprite;
	protected boolean				cDisableRemoveSprite;
	
	protected GProperties			cProperties;
	protected int[][]				cTile;
	protected GSprite[]				cSprite;
	protected int					cStartScrollX; // pixels
	protected int					cStartScrollY; // pixels
	
	
	public GLevel() {
		cDisableAddSprite = false;
		cDisableRemoveSprite = false;
		clear();
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		if (input == null)
			throw new IOException("GLevel.readFromFile: invalid file");
		
		version = input.pullInt(8);
		
		if (version == 1) {
			int[]					tileMapOneDimension;
			int						tileIndex;
			int						width;
			int						height;
			int						spriteCount;
			
			cProperties.readFromFile(input);
			
			width = input.pullInt(12);
			height = input.pullInt(12);
			
			tileMapOneDimension = input.decompress(width * height, 16);
			
			// convert tileMapOneDimension to cTile
			cTile = new int[width][height];
			tileIndex = 0;
			for (int y = 0; y < height; y++)
				for (int x = 0; x < width; x++)
					cTile[x][y] = tileMapOneDimension[tileIndex++];
			
			cStartScrollX = input.pullInt(16);
			cStartScrollY = input.pullInt(16);
			
			// read sprites
			spriteCount = input.pullInt(12);
			cSprite = new GSprite[spriteCount];
			for (int spriteCounter = 0; spriteCounter < spriteCount; spriteCounter++) {
				cSprite[spriteCounter] = GFactory.createSprite(input.pullInt(8), input);
				cSprite[spriteCounter].setLevel(this);
			}
		} else {
			throw new IOException("GLevel.readFromFile: Unknown version: " + version);
		}
	}
	
	public Object clone() {
		GLevel						other = null;
		
		try { other = (GLevel) super.clone(); } catch (CloneNotSupportedException e) {}
		
		// deep clone of properties
		other.cProperties = (GProperties) cProperties.clone();
		
		// deep clone of tiles
		other.cTile = new int[getWidth()][getHeight()];
		for (int x = 0; x < cTile.length; x++)
			System.arraycopy(cTile[x], 0, other.cTile[x], 0, cTile[x].length);
		
		// deep clone of sprites
		other.cSprite = new GSprite[cSprite.length];
		for (int index = 0; index < cSprite.length; index++) {
			other.cSprite[index] = (GSprite) cSprite[index].clone();
			other.cSprite[index].setLevel(other);
		}
		
		return other;
	}
	
	public void clear() {
		cProperties = GFactory.createProperties();
		cProperties.clear(sNumberCount, sFlagCount, sTextCount);
		
		cTile = new int[32][24];
		
		cStartScrollX = 0;
		cStartScrollY = 0;
		
		cSprite = new GSprite[0];
	}
	
	/**
	 * Updates the sprites, including moving and collision detection.
	 */
	public void nextFrame(int frame) {
		if (getSpriteCount() > 0) {
			int						spriteCounter;
			int						collideSpriteCounter;
			
			// pre-physics processing
			cDisableRemoveSprite = true;
			for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++)
				cSprite[spriteCounter].preProcess(frame);
			cDisableAddSprite = true;
			
			// move the sprites
			for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++)
				cSprite[spriteCounter].move(frame);
			
			// remove sprites outside of level area
			cDisableRemoveSprite = false;
			removeInvalidSprites();
			cDisableRemoveSprite = true;
			
			// detect sprite collisions by seeing if their collision rects overlap
			// each pair of sprites will only be evaluated once
			for (spriteCounter = 1; spriteCounter < getSpriteCount(); spriteCounter++)
				for (collideSpriteCounter = 0; collideSpriteCounter < spriteCounter; collideSpriteCounter++) {
					GSprite				one;
					GSprite				other;
					
					one = cSprite[spriteCounter];
					other = cSprite[collideSpriteCounter];
					
					if ((one.getCollisionRight() > other.getCollisionLeft()) &&
							(one.getCollisionBottom() > other.getCollisionTop()) &&
							(other.getCollisionRight() > one.getCollisionLeft()) &&
							(other.getCollisionBottom() > one.getCollisionTop())) {
						
						if (one.getDetectCollision())
							one.collideWith(other, frame);
						
						if (other.getDetectCollision())
							other.collideWith(one, frame);
					}
				}
			
			// post-physics processing
			cDisableRemoveSprite = false;
			for (spriteCounter = getSpriteCount() - 1; spriteCounter >= 0; spriteCounter--)
				cSprite[spriteCounter].postProcess(frame);
			cDisableAddSprite = false;
		}
	}
	
	public int getTile(int x, int y) {
		if (x < 0 || x >= getWidth()) {
			GOutput.throwOB("x", x, 0, getWidth() - 1);
			x = 0;
		}
		if (y < 0 || y >= getHeight()) {
			GOutput.throwOB("y", y, 0, getHeight() - 1);
			y = 0;
		}
		
		return cTile[x][y];
	}
	
	public void setTile(int x, int y, int tile) {
		if (x < 0 || x >= getWidth()) {
			GOutput.throwOB("x", x, 0, getWidth() - 1);
			return;
		}
		
		if (y < 0 || y >= getHeight()) {
			GOutput.throwOB("y", y, 0, getHeight() - 1);
			return;
		}
		
		cTile[x][y] = tile;
	}
	
	/**
	 * Returns the sprite with the specified index, or
	 * null if there is no sprite with the specified index.
	 */
	public GSprite getSprite(int index) {
		if (getSpriteCount() == 0) {
			GOutput.throwRuntime("getSprite: spriteCount = 0");
			return null;
		}
		
		if (index < 0 || index >= getSpriteCount()) {
			GOutput.throwOB("index", index, 0, getSpriteCount() - 1);
			return null;
		}
		
		return cSprite[index];
	}
	
	/**
	 * Returns the index of the sprite closest to the specified
	 * location, or -1 if the level has no sprites.
	 */
	public int getSpriteIndex(int locationX, int locationY) {
		double						candidate;
		double						closestDistanceSquared;
		int							closestIndex;
		int							spriteCounter;
		
		if (getSpriteCount() <= 0)
			return -1;
		
		// default to first sprite
		closestIndex = 0;
		closestDistanceSquared = Math.pow(cSprite[0].getLocationX() - locationX, 2) + 
					Math.pow(cSprite[0].getLocationY() - locationY, 2);
		
		// check all other sprites
		for (spriteCounter = 1; spriteCounter < getSpriteCount(); spriteCounter++) {
			candidate = Math.pow(cSprite[spriteCounter].getLocationX() - locationX, 2) + 
					Math.pow(cSprite[spriteCounter].getLocationY() - locationY, 2);
			if (candidate < closestDistanceSquared) {
				closestDistanceSquared = candidate;
				closestIndex = spriteCounter;
			}
		}
		
		return closestIndex;
	}
	
	/**
	 * Returns the sprite closest to the specified location, or
	 * null if the level has no sprites.
	 */
	public GSprite getSprite(int locationX, int locationY) {
		if (getSpriteCount() <= 0)
			return null;
		
		return cSprite[getSpriteIndex(locationX, locationY)];
	}
	
	/**
	 * Adds a sprite with the specified form index at the
	 * specified location, and initializes the sprite. Returns
	 * the newly created sprite.
	 */
	public GSprite addSprite(int x, int y, int formIndex) {
		GSprite[]					old;
		int							spriteCounter;
		
		if (cDisableAddSprite) {
			GOutput.throwRuntime("GSprite.addSprite(): not allowed during postProcess() and collideWith().");
			return null;
		}
		
		old = cSprite;
		cSprite = new GSprite[old.length + 1];
		for (spriteCounter = 0; spriteCounter < old.length; spriteCounter++)
			cSprite[spriteCounter] = old[spriteCounter];
		cSprite[old.length] = GFactory.createSprite(formIndex);
		cSprite[old.length].setLevel(this);
		cSprite[old.length].initialize(x, y);
		
		return cSprite[old.length];
	}
	
	/**
	 * Removes all sprites that have moved completely out of
	 * bounds.
	 */
	public void removeInvalidSprites() {
		int							spriteCounter;
		
		if (cDisableRemoveSprite) {
			GOutput.throwRuntime("GSprite.removeInvalidSprites(): not allowed during preProcess() and collideWith().");
			return;
		}
		
		spriteCounter = 0;
		while (spriteCounter < getSpriteCount()) {
			if (cSprite[spriteCounter].getRight() < 0 ||
					cSprite[spriteCounter].getBottom() < 0 ||
					cSprite[spriteCounter].getLeft() > (getWidth() * GTileForm.getTileWidth() << 10) ||
					cSprite[spriteCounter].getTop() > (getHeight() * GTileForm.getTileHeight() << 10))
				removeSprite(spriteCounter);
			else
				spriteCounter++;
		}
	}
	
	/**
	 * Removes the sprite with the specified index.
	 */
	public void removeSprite(int index) {
		GSprite[]					old;
		int							spriteCounter;
		
		if (cDisableRemoveSprite) {
			GOutput.throwRuntime("GSprite.removeSprite(): not allowed during preProcess() and collideWith().");
			return;
		}
		
		if (index < 0 || index >= getSpriteCount()) {
			GOutput.throwOB("index", index, 0, getSpriteCount() - 1);
			return;
		}
		
		cSprite[index].setLevel(null);
		
		if (getSpriteCount() == 1) {
			cSprite = new GSprite[0];
			return;
		}
		
		old = cSprite;
		cSprite = new GSprite[old.length - 1];
		if (index > 0)
			for (spriteCounter = 0; spriteCounter < index; spriteCounter++)
				cSprite[spriteCounter] = old[spriteCounter];
		if (index < cSprite.length)
			for (spriteCounter = index; spriteCounter < cSprite.length; spriteCounter++)
				cSprite[spriteCounter] = old[spriteCounter + 1];
	}
	
	/**
	 * Removes the specified sprite.
	 */
	public void removeSprite(GSprite sprite) {
		int							spriteCounter;
		
		spriteCounter = 0;
		while (spriteCounter < getSpriteCount()) {
			if (sprite == getSprite(spriteCounter)) {
				removeSprite(spriteCounter);
				return;
			} else
				spriteCounter++;
		}
	}
	
	public GProperties getProperties() { return cProperties; }
	public int getNumber(int index) { return cProperties.getNumber(index); }
	public boolean getFlag(int index) { return cProperties.getFlag(index); }
	public String getText(int index) { return cProperties.getText(index); }
	
	public int getWidth() { return cTile.length; }
	public int getHeight() { return cTile[0].length; }
	
	public int getStartScrollX() { return cStartScrollX; }
	public void setStartScrollX(int startScrollX) { cStartScrollX = startScrollX; }
	public int getStartScrollY() { return cStartScrollY; }
	public void setStartScrollY(int startScrollY) { cStartScrollY = startScrollY; }
	
	public int getSpriteCount() { return cSprite.length; }
	
	public void updateSpriteInstanceStatCount(int type) {
		int							spriteCounter;
		
		if (getSpriteCount() > 0)
			for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++)
				getSprite(spriteCounter).updateStatCount(type);
	}
	
	public static int getStaticNumberCount() { return sNumberCount; }
	
	public static void setStaticNumberCount(int numberCount) {
		int							levelCounter;
		
		if (numberCount < 0 || numberCount > GProperties.kMaxCount) {
			GOutput.throwOB("numberCount", numberCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sNumberCount = numberCount;
		
		for (levelCounter = 0; levelCounter < getCount(); levelCounter++)
			sLevel[levelCounter].getProperties().setNumberCount(sNumberCount);
	}
	
	public static int getStaticFlagCount() { return sFlagCount; }
	
	public static void setStaticFlagCount(int flagCount) {
		int							levelCounter;
		
		if (flagCount < 0 || flagCount > GProperties.kMaxCount) {
			GOutput.throwOB("flagCount", flagCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sFlagCount = flagCount;
		
		for (levelCounter = 0; levelCounter < getCount(); levelCounter++)
			sLevel[levelCounter].getProperties().setFlagCount(sFlagCount);
	}
	
	public static int getStaticTextCount() { return sTextCount; }
	
	public static void setStaticTextCount(int textCount) {
		int							levelCounter;
		
		if (textCount < 0 || textCount > GProperties.kMaxCount) {
			GOutput.throwOB("textCount", textCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sTextCount = textCount;
		
		for (levelCounter = 0; levelCounter < getCount(); levelCounter++)
			sLevel[levelCounter].getProperties().setTextCount(sTextCount);
	}
	
	public static void readAllFromFile(GBitReader input) throws IOException {
		int							version;
		int							levelCount;
		int							levelCounter;
		
		GOutput.logln("reading levels");
		
		if (input == null)
			throw new IOException("GLevel.readAllFromFile: invalid file");
		
		version = input.pullInt(8);
		
		GOutput.logln("version " + version);
		
		if (version == 1) {
			GOutput.log("id ");
			
			sNumberCount = input.pullInt(8);
			sFlagCount = input.pullInt(8);
			sTextCount = input.pullInt(8);
			
			levelCount = input.pullInt(8);
			if (levelCount == 0)
				throw new IOException("GLevel.readAllFromFile: invalid file format");
			
			sLevel = new GLevel[levelCount];
			for (levelCounter = 0; levelCounter < levelCount; levelCounter++) {
				GOutput.log(levelCounter + " ");
				sLevel[levelCounter] = GFactory.createLevel(input);
			}
			GOutput.logln("done");
		} else {
			throw new IOException("GLevel.readAllFromFile: unknown version: " + version);
		}
	}
	
	public static void clearAll() {
		sLevel = new GLevel[1];
		sLevel[0] = GFactory.createLevel();
	}
	
	public static int getCount() {
		if (sLevel == null)
			return 0;
		else
			return sLevel.length;
	}
	
	public static GLevel get(int index) {
		if (index < 0 || index >= getCount())
			GOutput.throwOB("index", index, 0, getCount() - 1);
		
		if (index < 0)
			index = 0;
		if (index > getCount() - 1)
			index = getCount() - 1;
		
		return sLevel[index];
	}
}

