package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;


/**
 * GLevelCanvas provides an efficient visual representation of a
 * level, drawing the tiles and sprites. It implements scrolling
 * by using a wrapped buffer, see the documentation for a discussion
 * on wrapped buffers.
 * <p>
 * The buffers are not guaranteed to be the same size as this
 * GLevelCanvas, see createBuffers() for details.
 * <p>
 * The first sprite is generally drawn first, so the last sprite
 * gets drawn on top of all other sprites. You can invoke
 * setReverseSpriteOrder(true) to reverse this drawing order,
 * and/or subclass GLevel to keep the sprites sorted any other
 * way.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GLevelCanvas
{
	// set to true to clear buffer between each update
	public boolean					cDebug;
	
	// currently displayed level, or null
	protected GLevel				cLevel;
	protected Component				cImageProducer;
	
	// stored locally to detect changes
	protected int					cTileWidth;
	protected int					cTileHeight;
	
	// requested buffer size
	protected int					cRequestedWidth;
	protected int					cRequestedHeight;
	
	// dimensions, in tiles and pixels
	protected int					cBufferTileWidth; // tiles
	protected int					cBufferTileHeight; // tiles
	protected int					cBufferWidth; // pixels
	protected int					cBufferHeight; // pixels
	
	// wrapped pixel buffer data
	protected int[]					cBufferPixels;
	protected MemoryImageSource		cBufferSource;
	protected Image					cPixelBuffer;
	
	// non-wrapped assembly buffer data
	protected Image					cBuffer;
	protected Graphics				cBufferGraphics;
	
	// current state info
	protected int					cScrollX;
	protected int					cScrollY;
	protected int					cFrame;
	
	// keeping track of newly exposed tiles
	protected int					cLastUpdateScrollX;
	protected int					cLastUpdateScrollY;
	protected int					cLastUpdateFrame;
	protected boolean				cFreshBuffer;
	protected boolean[][]			cNeedUpdate;
	
	protected boolean				cReverseSpriteOrder; // paints last sprite first, first sprite on top
	
	
	/**
	 * The contructor requires an imageProducer to create its buffers,
	 * generally passing the applet is recommended. Of course,
	 * imageProducer must be able to produce images with the
	 * createImage() methods. As far as I know, any visible component
	 * will do.
	 * <p>
	 * level can be null, the buffers will still be created. Of course, this
	 * GLevelCanvas will need a invoke to setLevel() to display any level.
	 * <p>
	 * If width or height is less than or equal to zero, no buffer will
	 * be created.
	 */
	public GLevelCanvas(Component imageProducer, GLevel level, int width, int height) {
		if (imageProducer == null)
			throw new IllegalArgumentException("GLevelCanvas.GLevelCanvas: imageProducer is null");
		
		cImageProducer = imageProducer;
		cReverseSpriteOrder = false;
		
		cRequestedWidth = width;
		cRequestedHeight = height;
		
		cBufferWidth = -1;
		cBufferHeight = -1;
		
		setLevel(level);
	}
	
	public void setLevel(GLevel level) {
		cLevel = level;
		
		createBuffers();
		
		if (cLevel == null) {
			cScrollX = 0;
			cScrollY = 0;
			cFrame = 0;
		} else {
			setScrollX(cLevel.getStartScrollX());
			setScrollY(cLevel.getStartScrollY());
			
			cFrame = 0;
			cDebug = false;
			
			cLastUpdateScrollX = cScrollX;
			cLastUpdateScrollY = cScrollY;
			cLastUpdateFrame = 0;
			
			cFreshBuffer = true;
		}
	}
	
	/**
	 * If a level has been provided, the buffers will be trimmed to an
	 * even multiple of the tile width/height. They will also be
	 * trimmed to the size of the level, if the buffers are large.
	 * Also, if this GLevelCanvas is smaller than the tile size, the
	 * buffer will at least have the same size as a tile.
	 * <p>
	 * If no level (that is, null) has been provided, the buffers
	 * will be the same size as the GLevelCanvas.
	 */
	public void createBuffers() {
		boolean						reallocateBuffers;
		
		if (cRequestedWidth <= 0 || cRequestedHeight <= 0) {
			// release buffers
			cBufferWidth = 0;
			cBufferHeight = 0;
			cBufferTileWidth = 0;
			cBufferTileHeight = 0;
			
			cBufferPixels = null;
			cBufferSource = null;
			cPixelBuffer = null;
			cBuffer = null;
			cBufferGraphics = null;
			return;
		}
		
		if (cTileWidth != GTileForm.getTileWidth() || cTileHeight != GTileForm.getTileHeight()) {
			// the tile size has changed
			cTileWidth = GTileForm.getTileWidth();
			cTileHeight = GTileForm.getTileHeight();
			// invalidate buffers
			cBufferTileWidth = -1;
			cBufferTileHeight = -1;
			cBufferWidth = -1;
			cBufferHeight = -1;
		}
		
		if (cLevel == null) {
			// the buffer will be the same size as this GLevelCanvas
			
			cBufferTileWidth = -1;
			cBufferTileHeight = -1;
			cFreshBuffer = true;
			cNeedUpdate = null;
			cScrollX = 0;
			cScrollY = 0;
			
			if (cRequestedWidth != cBufferWidth || cRequestedHeight != cBufferHeight) {
				reallocateBuffers = true;
				cBufferWidth = cRequestedWidth;
				cBufferHeight = cRequestedHeight;
			} else
				reallocateBuffers = false; // same size
		} else {
			// the buffer will be trimmed to a multiple of the tile size
			int						newWidth;
			int						newHeight;
			
			// truncate to closest full tile
			newWidth = cRequestedWidth / cTileWidth;
			newHeight = cRequestedHeight / cTileHeight;
			
			// can not be smaller than a tile
			if (newWidth < 1)
				newWidth = 1;
			if (newHeight < 1)
				newHeight = 1;
			
			// can not be larger than the level
			if (newWidth > cLevel.getWidth())
				newWidth = cLevel.getWidth();
			if (newHeight > cLevel.getHeight())
				newHeight = cLevel.getHeight();
			
			if (newWidth != cBufferTileWidth || newHeight != cBufferTileHeight) {
				reallocateBuffers = true;
				
				cBufferTileWidth = newWidth;
				cBufferTileHeight = newHeight;
				
				cBufferWidth = cBufferTileWidth * cTileWidth;
				cBufferHeight = cBufferTileHeight * cTileHeight;
				
				// make sure we are not scrolled out of bounds
				setScrollX(cScrollX);
				setScrollY(cScrollY);
				
				cFreshBuffer = true;
				cNeedUpdate = new boolean[cBufferTileWidth][cBufferTileHeight];
			} else
				reallocateBuffers = false; // same size
		}
		
		if (reallocateBuffers) {
			cBufferPixels = new int[cBufferWidth * cBufferHeight];
			cBufferSource = new MemoryImageSource(cBufferWidth, cBufferHeight, cBufferPixels, 0, cBufferWidth);
			cBufferSource.setAnimated(true);
			
			cPixelBuffer = cImageProducer.createImage(cBufferSource);
			cBuffer = cImageProducer.createImage(cBufferWidth, cBufferHeight);
			
			if (cBuffer == null) {
				cBufferGraphics = null;
				throw new NullPointerException("GLevelCanvas.createBuffers: no double-buffer; invalid image producer: " + cImageProducer);
			} else
				cBufferGraphics = cBuffer.getGraphics();
		}
	}
	
	/**
	 * Changes the size of this GLevelCanvas, and updates the size
	 * of the buffers. The buffers are not guaranteed to be the same
	 * size as this GLevelCanvas, see createBuffers() for details.
	 * <p>
	 * If width or height is less than or equal to zero, no buffer will
	 * be created.
	 */
	public void setSize(int width, int height) {
		cRequestedWidth = width;
		cRequestedHeight = height;
		
		createBuffers();
	}
	
	/**
	 * Marks the tile at the specified coordinates as needing an
	 * update.
	 */
	public void setNeedUpdate(int x, int y) {
		int							localX;
		int							localY;
		
		if (cLevel == null)
			return; // no tiles to update
		
		if ((x + 1) * cTileWidth <= cScrollX ||
				(y + 1) * cTileHeight <= cScrollY ||
				(x - cBufferTileWidth) * cTileWidth > cScrollX ||
				(y - cBufferTileHeight) * cTileHeight > cScrollY)
			return; // the tile is not visible
		
		localX = x % cBufferTileWidth;
		localY = y % cBufferTileHeight;
		
		cNeedUpdate[localX][localY] = true;
	}
	
	/**
	 * Marks this GLevelCanvas as needing a complete update,
	 * or no update at all.
	 */
	public void setNeedUpdate(boolean needUpdate) {
		if (cLevel == null)
			; // no level to update
		else if (needUpdate)
			cFreshBuffer = true;
		else {
			cFreshBuffer = false;
			for (int x = 0; x < cBufferTileWidth; x++)
				for (int y = 0; y < cBufferTileHeight; y++)
					cNeedUpdate[x][y] = false;
		}
	}
	
	/**
	 * Paints the tiles that have been exposed by horizontal
	 * scrolling since the last update.
	 */
	protected void paintHorizontalUpdate() {
		int							breakX;
		int							breakY;
		int							topUpdate;
		int							bottomUpdate;
		int							plotTop;
		int							plotBottom;
		boolean						brokenX;
		int							globalX;
		int							globalY;
		int							tileX;
		int							tileY;
		
		breakX = cScrollX % cBufferWidth;
		breakY = cScrollY % cBufferHeight;
		
		if (cScrollY < cLastUpdateScrollY) {
			// scrolling up
			topUpdate = cScrollY % cBufferHeight;
			bottomUpdate = cLastUpdateScrollY % cBufferHeight;
			globalY = cScrollY / cTileHeight;
		} else {
			// scrolling down
			topUpdate = cLastUpdateScrollY % cBufferHeight;
			bottomUpdate = cScrollY % cBufferHeight;
			globalY = cLastUpdateScrollY / cTileHeight + cBufferTileHeight;
		}
		
		if (bottomUpdate < topUpdate)
			bottomUpdate += cBufferHeight;
		
		for (tileY = (topUpdate / cTileHeight); tileY <= (bottomUpdate / cTileHeight); tileY++) {
			if (globalY < 0 || globalY >= cLevel.getHeight())
				continue;
			
			if (tileY == (topUpdate / cTileHeight))
				plotTop = topUpdate % cTileHeight;
			else
				plotTop = 0;
			if (tileY == (bottomUpdate / cTileHeight))
				plotBottom = bottomUpdate % cTileHeight;
			else
				plotBottom = cTileHeight;
			
			for (tileX = 0; tileX < cBufferTileWidth; tileX++) {
				if (!cNeedUpdate[tileX][tileY % cBufferTileHeight]) { // if cNeedUpdate, paintChanged will update it
					if (tileX < breakX / cTileWidth) {
						// left wBuffer panel
						brokenX = false;
						globalX = tileX + ((cScrollX / cBufferWidth + 1) * cBufferTileWidth);
					}
					else if (tileX > breakX / cTileWidth) {
						// right wBuffer panel
						brokenX = false;
						globalX = tileX + ((cScrollX / cBufferWidth) * cBufferTileWidth);
					} else {
						// at border of panels
						brokenX = true;
						globalX = tileX + ((cScrollX / cBufferWidth) * cBufferTileWidth);
					}
					
					if (globalX < 0 || globalX >= cLevel.getWidth())
						continue;
					
					if (brokenX) {
						// relative left in wBuffer
						if (globalX + cBufferTileWidth < cLevel.getWidth())
							GTileForm.get(cLevel.getTile(globalX + cBufferTileWidth, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
									0, plotTop, cScrollX % cTileWidth, plotBottom, tileX * cTileWidth, (tileY % cBufferTileHeight) * cTileHeight);
						
						// relative bottom in wBuffer
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								cScrollX % cTileWidth, plotTop, cTileWidth, plotBottom, tileX * cTileWidth, (tileY % cBufferTileHeight) * cTileHeight);
					} else {
						// not split left/right
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								0, plotTop, cTileWidth, plotBottom, tileX * cTileWidth, (tileY % cBufferTileHeight) * cTileHeight);
					}
				}
			}
			globalY++;
		}
	}
	
	/**
	 * Paints the tiles that have been exposed by vertical
	 * scrolling since the last update.
	 */
	protected void paintVerticalUpdate() {
		int							breakX;
		int							breakY;
		int							leftUpdate;
		int							rightUpdate;
		int							plotLeft;
		int							plotRight;
		boolean						brokenY;
		int							globalX;
		int							globalY;
		int							tileX;
		int							tileY;
		
		breakX = cScrollX % cBufferWidth;
		breakY = cScrollY % cBufferHeight;
		
		if (cScrollX < cLastUpdateScrollX) {
			// scrolling left
			leftUpdate = cScrollX % cBufferWidth;
			rightUpdate = cLastUpdateScrollX % cBufferWidth;
			globalX = cScrollX / cTileWidth;
		} else {
			// scrolling right
			leftUpdate = cLastUpdateScrollX % cBufferWidth;
			rightUpdate = cScrollX % cBufferWidth;
			globalX = cLastUpdateScrollX / cTileWidth + cBufferTileWidth;
		}
		
		if (rightUpdate < leftUpdate)
			rightUpdate += cBufferWidth;
		
		for (tileX = (leftUpdate / cTileWidth); tileX <= (rightUpdate / cTileWidth); tileX++) {
			if (globalX < 0 || globalX >= cLevel.getWidth())
				continue;
			
			if (tileX == (leftUpdate / cTileWidth))
				plotLeft = leftUpdate % cTileWidth;
			else
				plotLeft = 0;
			if (tileX == (rightUpdate / cTileWidth))
				plotRight = rightUpdate % cTileWidth;
			else
				plotRight = cTileWidth;
			
			for (tileY = 0; tileY < cBufferTileHeight; tileY++) {
				if (!cNeedUpdate[tileX % cBufferTileWidth][tileY]) { // if cNeedUpdate, paintChanged will update it
					if (tileY < breakY / cTileHeight) {
						// top wBuffer panel
						brokenY = false;
						globalY = tileY + ((cScrollY / cBufferHeight + 1) * cBufferTileHeight);
					}
					else if (tileY > breakY / cTileHeight) {
						// bottom wBuffer panel
						brokenY = false;
						globalY = tileY + ((cScrollY / cBufferHeight) * cBufferTileHeight);
					} else {
						// at border of panels
						brokenY = true;
						globalY = tileY + ((cScrollY / cBufferHeight) * cBufferTileHeight);
					}
					
					if (globalY < 0 || globalY >= cLevel.getHeight())
						continue;
					
					if (brokenY) {
						// relative top in wBuffer
						if (globalY + cBufferTileHeight < cLevel.getHeight())
							GTileForm.get(cLevel.getTile(globalX, globalY + cBufferTileHeight)).draw(cFrame, cBufferPixels, cBufferWidth,
									plotLeft, 0, plotRight, cScrollY % cTileHeight, (tileX % cBufferTileWidth) * cTileWidth, tileY * cTileHeight);
						
						// relative bottom in wBuffer
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								plotLeft, cScrollY % cTileHeight, plotRight, cTileHeight, (tileX % cBufferTileWidth) * cTileWidth, tileY * cTileHeight);
					} else {
						// not split top/bottom
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								plotLeft, 0, plotRight, cTileHeight, (tileX % cBufferTileWidth) * cTileWidth, tileY * cTileHeight);
					}
				}
			}
			globalX++;
		}
	}
	
	/**
	 * Paints the animated tiles that have changed view
	 * since the last update, and tiles that were covered
	 * by sprites last update (and any other sprites marked
	 * as needing updating).
	 */
	protected void paintChanged() {
		int							tileX;
		int							tileY;
		int							breakX;
		int							breakY;
		
		breakX = cScrollX % cBufferWidth;
		breakY = cScrollY % cBufferHeight;
		
		for (tileY = 0; tileY < cBufferTileHeight; tileY++) {
			int						globalY;
			boolean					brokenY;
			
			if (tileY < breakY / cTileHeight) {
				// top wBuffer panel
				brokenY = false;
				globalY = tileY + ((cScrollY / cBufferHeight + 1) * cBufferTileHeight);
			}
			else if (tileY > breakY / cTileHeight) {
				// bottom wBuffer panel
				brokenY = false;
				globalY = tileY + ((cScrollY / cBufferHeight) * cBufferTileHeight);
			} else {
				// at border of panels
				brokenY = true;
				// globalY is only part of the tile
				globalY = tileY + ((cScrollY / cBufferHeight) * cBufferTileHeight);
			}
			
			if (globalY < 0 || globalY >= cLevel.getHeight())
				continue;
			
			for (tileX = 0; tileX < cBufferTileWidth; tileX++) {
				int					globalX;
				boolean				brokenX;
				
				if (tileX < breakX / cTileWidth) {
					// left wBuffer panel
					brokenX = false;
					globalX = tileX + ((cScrollX / cBufferWidth + 1) * cBufferTileWidth);
				}
				else if (tileX > breakX / cTileWidth) {
					// right wBuffer panel
					brokenX = false;
					globalX = tileX + ((cScrollX / cBufferWidth) * cBufferTileWidth);
				} else {
					// at border of panels
					brokenX = true;
					// globalX is only part of the tile
					globalX = tileX + ((cScrollX / cBufferWidth) * cBufferTileWidth);
				}
				
				if (globalX < 0 || globalX >= cLevel.getWidth())
					continue;
				
				if (cNeedUpdate[tileX][tileY] || cFreshBuffer) {
					if (brokenY && brokenX) {
						// relative top-left in wBuffer
						if (globalY + cBufferTileHeight < cLevel.getHeight())
							if (globalX + cBufferTileWidth < cLevel.getWidth())
								GTileForm.get(cLevel.getTile(globalX + cBufferTileWidth, globalY + cBufferTileHeight)).draw(cFrame, cBufferPixels, cBufferWidth,
										0, 0, cScrollX % cTileWidth, cScrollY % cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
						
						// relative bottom-left in wBuffer
						if (globalX + cBufferTileWidth < cLevel.getWidth())
							GTileForm.get(cLevel.getTile(globalX + cBufferTileWidth, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
									0, cScrollY % cTileHeight, cScrollX % cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
						
						// relative top-right in wBuffer
						if (globalY + cBufferTileHeight < cLevel.getHeight())
							GTileForm.get(cLevel.getTile(globalX, globalY + cBufferTileHeight)).draw(cFrame, cBufferPixels, cBufferWidth,
									cScrollX % cTileWidth, 0, cTileWidth, cScrollY % cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
						
						// relative bottom-right in wBuffer
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								cScrollX % cTileWidth, cScrollY % cTileHeight, cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
					}
					else if (brokenY) {
						// relative top in wBuffer
						if (globalY + cBufferTileHeight < cLevel.getHeight())
							GTileForm.get(cLevel.getTile(globalX, globalY + cBufferTileHeight)).draw(cFrame, cBufferPixels, cBufferWidth,
									0, 0, cTileWidth, cScrollY % cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
						
						// relative bottom in wBuffer
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								0, cScrollY % cTileHeight, cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
					}
					else if (brokenX) {
						// relative left in wBuffer
						if (globalX + cBufferTileWidth < cLevel.getWidth())
							GTileForm.get(cLevel.getTile(globalX + cBufferTileWidth, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
									0, 0, cScrollX % cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
						
						// relative right in wBuffer
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								cScrollX % cTileWidth, 0, cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
					} else {
						GTileForm.get(cLevel.getTile(globalX, globalY)).draw(cFrame, cBufferPixels, cBufferWidth,
								0, 0, cTileWidth, cTileHeight, tileX * cTileWidth, tileY * cTileHeight);
					}
					
					cNeedUpdate[tileX][tileY] = false;
				}
			}
		}
	}
	
	/**
	 * Draws all sprites, and marks the tiles below them as needing
	 * to be updated the next frame.
	 */
	protected void paintSprites() {
		int							spriteCounter;
		
		if (cReverseSpriteOrder)
			spriteCounter = cLevel.getSpriteCount() - 1;
		else
			spriteCounter = 0;
		
		while (spriteCounter < cLevel.getSpriteCount() && spriteCounter >= 0) {
			GSprite					theSprite;
			int						x;
			int						y;
			
			theSprite = cLevel.getSprite(spriteCounter);
			if (theSprite == null) {
				// should never be here
			} else if ((theSprite.getRight() >> 10) < cScrollX ||
				(theSprite.getBottom() >> 10) < cScrollY ||
				(theSprite.getLeft() >> 10) > cScrollX + cBufferWidth ||
				(theSprite.getTop() >> 10) > cScrollY + cBufferHeight) {
				// sprite is not visible
			} else {
				// make sure the tiles behind the sprite get updated next frame
				for (x = (theSprite.getLeft() >> 10) / cTileWidth;
						x <= (((theSprite.getRight() >> 10) - 1) / cTileWidth); x++)
					for (y = (theSprite.getTop() >> 10) / cTileHeight;
							y <= (((theSprite.getBottom() >> 10) - 1) / cTileHeight); y++)
						if (x >= 0 && y >= 0)
							setNeedUpdate(x, y); // will trim to the buffer size
				
				theSprite.draw(cBufferPixels, cBufferWidth, cScrollX, cScrollY);
			}
			
			if (cReverseSpriteOrder)
				spriteCounter--;
			else
				spriteCounter++;
		}
	}
	
	public void paint(Graphics finalGraphics) {
		paint(finalGraphics, cBufferGraphics, cFrame);
	}
	
	public void paint(Graphics finalGraphics, int frame) {
		paint(finalGraphics, cBufferGraphics, frame);
	}
	
	/**
	 * Updates the buffer, including new areas exposed by scrolling,
	 * tiles previously covered by sprites, animated tiles and
	 * all sprites.
	 */
	public void paint(Graphics finalGraphics, Graphics midGraphics, int frame) {
		int							breakX;
		int							breakY;
		
		if (!preparePaint())
			return;
		
		cFrame = frame;
		
		breakX = cScrollX % cBufferWidth;
		breakY = cScrollY % cBufferHeight;
		
		if (cDebug) {
			// fill the entire buffer with grey, helps seeing how much
			// is updated each frame
			for (int index = 0; index < cBufferPixels.length; index++)
				cBufferPixels[index] = 0xFF7F7F7F;
		}
		
		// if update area is significant update all
		if (Math.abs(cScrollX - cLastUpdateScrollX) * 3 > cBufferWidth * 2)
			cFreshBuffer = true;
		if (Math.abs(cScrollY - cLastUpdateScrollY) * 3 > cBufferHeight * 2)
			cFreshBuffer = true;
		
			// draw tiles
		if (cFreshBuffer) {
			// if the buffer is completely fresh, everything needs to be
			// drawn, and all tiles are considered changed
			paintChanged();
			cFreshBuffer = false;
		} else {
			// make sure animated tiles are updated
			if (cFrame != cLastUpdateFrame) {
				int					x;
				int					y;
				int					animationPeriod;
				
				// this will unfortunately update an entire tile when part of it is animated
				for (x = cScrollX / cTileWidth; x <= (cScrollX / cTileWidth + cBufferTileWidth); x++)
					if (x >= 0 && x < cLevel.getWidth())
						for (y = cScrollY / cTileHeight; y <= (cScrollY / cTileHeight + cBufferTileHeight); y++)
							if (y >= 0 && y < cLevel.getHeight()) {
								animationPeriod = GTileForm.get(cLevel.getTile(x, y)).getAnimationPeriod();
								if (animationPeriod > 0 && (cFrame / animationPeriod) != (cLastUpdateFrame / animationPeriod))
									cNeedUpdate[x % cBufferTileWidth][y % cBufferTileHeight] = true;
							}
			}
			
			/*
			Peformance note (to avoid unnecessary drawing):
			paintVerticalUpdate and paintHorizontalUpdate should be invoked
			before any paintChanged is invoked, because they will not draw any
			tile that is marked as cNeedUpdate, instead leaving it
			to be updated by paintChanged.
			*/
			
			if (cScrollX != cLastUpdateScrollX) {
				// update between any left and right panel
				paintVerticalUpdate();
			}
			
			if (cScrollY != cLastUpdateScrollY) {
				// update between any top and bottom panel
				paintHorizontalUpdate();
			}
			
			// changed generally means previously covered by sprites, or animated
			paintChanged();
		}
		
		// draw sprites
		if (cLevel.getSpriteCount() > 0)
			paintSprites();
		
		cBufferSource.newPixels();
		cLastUpdateScrollX = cScrollX;
		cLastUpdateScrollY = cScrollY;
		cLastUpdateFrame = cFrame;
		
		// assemble wrapped buffer in a non-wrapped buffer
		copyBuffers(finalGraphics, midGraphics, breakX, breakY);
	}
	
	/**
	 * Returns true if this GLevelCanvas is ready to be painted.
	 */
	public boolean preparePaint() {
		if (cLevel == null) {
			GOutput.throwRuntime("GLevelCanvas.paint(): cLevel == null");
			return false;
		} else if (cBuffer == null || cPixelBuffer == null) {
			GOutput.throwRuntime("GLevelCanvas.paint(): no buffer");
			return false;
		}
		
		if (cTileWidth != GTileForm.getTileWidth() || cTileHeight != GTileForm.getTileHeight()) {
			// new tile size
			createBuffers();
		}
		
		return true;
	}
	
	public void copyBuffers(Graphics finalGraphics, Graphics midGraphics, int breakX, int breakY) {
		Shape						originalMidClip = null;
		Shape						originalFinalClip = null;
		
		
		if (midGraphics != null)
			originalMidClip = midGraphics.getClip();
		if (finalGraphics != null)
			originalFinalClip = finalGraphics.getClip();
		
		if (midGraphics != null) {
			// copy buffer to midbuffer
			midGraphics.setClip(0, 0, cBufferWidth - breakX, cBufferHeight - breakY);
			midGraphics.drawImage(cPixelBuffer, -breakX, -breakY, null);
			
			midGraphics.setClip(cBufferWidth - breakX, 0, breakX, cBufferHeight - breakY);
			midGraphics.drawImage(cPixelBuffer, cBufferWidth - breakX, -breakY, null);
			
			midGraphics.setClip(0, cBufferHeight - breakY, cBufferWidth - breakX, breakY);
			midGraphics.drawImage(cPixelBuffer, -breakX, cBufferHeight - breakY, null);
			
			midGraphics.setClip(cBufferWidth - breakX, cBufferHeight - breakY, breakX, breakY);
			midGraphics.drawImage(cPixelBuffer, cBufferWidth - breakX, cBufferHeight - breakY, null);
			
			if ((midGraphics == cBufferGraphics) && (finalGraphics != null)) {
				// copy midbuffer to window
				finalGraphics.drawImage(cBuffer, 0, 0, null);
			}
		} else if (finalGraphics != null) {
			// the mid (buffer) graphics does not exist
			// copy buffer to window
			finalGraphics.setClip(0, 0, cBufferWidth - breakX, cBufferHeight - breakY);
			finalGraphics.drawImage(cPixelBuffer, -breakX, -breakY, null);
			
			finalGraphics.setClip(cBufferWidth - breakX, 0, breakX, cBufferHeight - breakY);
			finalGraphics.drawImage(cPixelBuffer, cBufferWidth - breakX, -breakY, null);
			
			finalGraphics.setClip(0, cBufferHeight - breakY, cBufferWidth - breakX, breakY);
			finalGraphics.drawImage(cPixelBuffer, -breakX, cBufferHeight - breakY, null);
			
			finalGraphics.setClip(cBufferWidth - breakX, cBufferHeight - breakY, breakX, breakY);
			finalGraphics.drawImage(cPixelBuffer, cBufferWidth - breakX, cBufferHeight - breakY, null);
			
		}
		
		if (midGraphics != null)
			midGraphics.setClip(originalMidClip);
		if (finalGraphics != null)
			finalGraphics.setClip(originalFinalClip);
		
		// copy buffer to window
		// mostly useful for debugging
		// finalGraphics.drawImage(cPixelBuffer, 0, 0, null);
	}
	
	/**
	 * Returns the simple buffer, not the pixel buffer.
	 */
	public Image getBuffer() {
		return cBuffer;
	}
	
	/**
	 * Returns the pixel buffer.
	 */
	public Image getPixelBuffer() {
		return cPixelBuffer;
	}
	
	public int getFrame() {
		return cFrame;
	}
	
	/**
	 * Reverses the order in which the sprites are drawn, and
	 * which sprites appear in front of others.
	 * <ul>
	 * <li>If reverse is false, the first sprite is drawn first,
	 * and the last sprite appears on top.
	 * <li>If reverse is true, the last sprite is drawn first,
	 * and the first sprite appears on top.
	 * </ul>
	 * For more precise ordering, you can subclass GLevel to keep the
	 * sprites sorted any way you like.
	 */
	public void setReverseSpriteOrder(boolean reverse) {
		cReverseSpriteOrder = reverse;
	}
	
	public int getScrollX() {
		return cScrollX;
	}
	
	public int getScrollY() {
		return cScrollY;
	}
	
	public void setScrollX(int scrollX) {
		int							newScrollX;
		
		newScrollX = scrollX;
		if (cLevel != null && newScrollX > (cLevel.getWidth() * cTileWidth - cBufferWidth))
			newScrollX = cLevel.getWidth() * cTileWidth - cBufferWidth;
		if (newScrollX < 0)
			newScrollX = 0;
		
		cScrollX = newScrollX;
	}
	
	public void setScrollY(int scrollY) {
		int							newScrollY;
		
		newScrollY = scrollY;
		if (cLevel != null && newScrollY > (cLevel.getHeight() * cTileHeight - cBufferHeight))
			newScrollY = cLevel.getHeight() * cTileHeight - cBufferHeight;
		if (newScrollY < 0)
			newScrollY = 0;
		
		cScrollY = newScrollY;
	}
}

