package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GOutput controls certain messages that the GameForm classes
 * send; depending on the settings, the same message will
 * either generate an exception, a printed message (with or without
 * a stack trace) or do nothing.
 * <p>
 * Checked exceptions (notably IOExceptions) will be thrown by
 * the GameForm classes, but non-fatal exceptions like
 * IllegalStateExceptions and IllegalArgumentException will be
 * passed to GOutput. If sThrowExceptions is true, an exception
 * will be thrown, otherwise a message will be printed and
 * execution will continue. During development and debugging,
 * exceptions are generally helpful, but you may want to hide
 * them (and hope they will pass unnoticed) in the final product.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GOutput
{
	/**
	 * If sMute is true, any passed messages will not be shown,
	 * nor will exceptions be thrown.
	 */
	public static boolean			sMute = false;
	
	/**
	 * Determines whether to throw exceptions, or print a message.
	 */
	public static boolean			sThrowExceptions = true;
	
	/**
	 * Determines whether, if sThrowExceptions is false, to print
	 * a stack trace with the error description.
	 */
	public static boolean			sTraceStack = true;
	
	protected static PrintStream	sOutputStream = System.out;
	
	
	/**
	 * Assign a different output stream to receive the messages.
	 */
	public static void setOutputStream(PrintStream output) {
		if (output != null)
			sOutputStream = output;
		else
			sOutputStream = System.out;
	}
	
	/**
	 * Prints a message to the output stream.
	 */
	public static void logln(String message) {
		log(message + "\n");
	}
	
	/**
	 * Prints a message to the output stream.
	 */
	public static void log(Object message) {
		if (!sMute)
			sOutputStream.print(message.toString());
	}
	
	/**
	 * Signals that a runtime exception should be thrown and
	 * handled.
	 */
	public static void throwRuntime(String text) {
		if (sMute)
			return;
		
		if (sThrowExceptions)
			throw new RuntimeException(text);
		else {
			logln(text);
			if (sTraceStack)
				(new Throwable()).printStackTrace(sOutputStream);
		}
	}
	
	/**
	 * Signals that a method argument is out of bounds.
	 */
	public static void throwOB(String name, int value, int min, int max) {
		if (sMute)
			return;
		
		if (sThrowExceptions) {
			if (min <= max)
				throw new IllegalArgumentException(name + " == " + value + "; valid values: " + min + "-" + max);
			else
				throw new IllegalStateException(name + " == " + value + "; no valid values");
		} else {
			if (min <= max)
				logln("Illegal argument: " + name + " == " + value + "; valid values: " + min + "-" + max);
			else
				logln("Illegal state: " + name + " == " + value + "; no valid values");
			
			if (sTraceStack)
				(new Throwable()).printStackTrace(sOutputStream);
		}
	}
}

