package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GPixels is GameForm's image format. It can be drawn to a pixel
 * buffer, in any of eight modes (flipped and rotated). It only
 * supports 32-bit argb, including full 8-bit mask support.
 * <p>
 * GPixels can be drawn to a pixel buffer (using draw()), or to
 * a wrapped pixel buffer (using the methods drawTile() or
 * drawWrapped()). See the documentation for a discussion on
 * wrapped buffers.
 * <p>
 * The different modes apply transformations to the image when
 * drawing it.
 * <ul>
 * <li> mode 0: no modification
 * <li> mode 1: rotate 90 degrees clockwise.
 * <li> mode 2: rotate 180 degrees clockwise.
 * <li> mode 3: rotate 270 degrees clockwise.
 * <li> mode 4: flip horizontally.
 * <li> mode 5: rotate 90 degrees clockwise, then flip horizontally.
 * <li> mode 6: rotate 180 degrees clockwise, then flip horizontally.
 * <li> mode 7: rotate 270 degrees clockwise, then flip horizontally.
 * </ul>
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GPixels implements Cloneable
{
	public static final int			kMaxSize = 1023; // for storing size in 10 bits
	
	protected int					cWidth;
	protected int					cHeight;
	protected int[]					cPixelArray;
	protected int					cAlphaMod; // modifies transparency, not saved
	protected boolean				cHasMask;
	
	
	public GPixels() {
		cAlphaMod = 255;
		setSize(GTileForm.getTileWidth(), GTileForm.getTileHeight());
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							width;
		int							height;
		int							version;
		
		if (input == null)
			throw new IOException("GPixels.readFromFile: invalid file");
		
		cAlphaMod = 255; // not saved
		
		version = input.pullInt(8);
		
		if (version == 1) {
			width = input.pullInt(10);
			height = input.pullInt(10);
			
			setSize(width, height);
			
			cHasMask = input.pullBoolean(); // compression info only
			if (cHasMask) {
				cPixelArray = input.decompress(width * height, 32);
			} else {
				// use only 24 low bits
				cPixelArray = input.decompress(width * height, 24);
				
				for (int pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++)
					cPixelArray[pixelIndex] |= 0xFF000000;
			}
		} else {
			throw new IOException("GPixels.readFromFile: Unknown version: " + version);
		}
	}
	
	public Object clone() {
		GPixels						other = null;
		
		try { other = (GPixels) super.clone(); } catch (CloneNotSupportedException e) {}
		
		// deep clone of pixel array
		other.cPixelArray = new int[cPixelArray.length];
		System.arraycopy(cPixelArray, 0, other.cPixelArray, 0, cPixelArray.length);
		
		return other;
	}
	
	public void clear() {
		for (int pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++)
			cPixelArray[pixelIndex] = 0xFFFFFFFF;
	}
	
	/**
	 * Sets the size of the image, destroying any previous image data.
	 */
	public void setSize(int width, int height) {
		if (width < 1 || width > kMaxSize) {
			GOutput.throwOB("width", width, 1, kMaxSize);
			return;
		}
		
		if (height < 1 || height > kMaxSize) {
			GOutput.throwOB("height", height, 1, kMaxSize);
			return;
		}
		
		cWidth = width;
		cHeight = height;
		
		cPixelArray = new int[cWidth * cHeight];
		clear();
	}
	
	/**
	 * Accepts clipped dimensions only, supports any mask and any mode.
	 */
	protected void copyPixelsAlpha(int[] canvas, int canvasWidth, int sourceStart, int destStart,
			int plotWidth, int plotHeight, int deltaX, int deltaY) {
		
		int							yCounter;
		int							xCounter;
		int							sourceIndex;
		int							destIndex;
		
		int							sourceAlpha;
		int							destAlpha;
		int							red;
		int							green;
		int							blue;
		
		sourceIndex = sourceStart;
		destIndex = destStart;
		
		for (yCounter = 0; yCounter < plotHeight; yCounter++) {
			for (xCounter = 0; xCounter < plotWidth; xCounter++) {
				sourceAlpha = ((cPixelArray[sourceIndex] >> 24) & 0xFF) * cAlphaMod / 255;
				
				if (sourceAlpha == 255) // opaque
					canvas[destIndex] = cPixelArray[sourceIndex];
				else if (sourceAlpha > 0) { // semi-transparent
					destAlpha = 0xFF - sourceAlpha;
					
					red = ((cPixelArray[sourceIndex] >> 16) & 0xFF) * sourceAlpha / 255 +
						((canvas[destIndex] >> 16) & 0xFF) * destAlpha / 255;
					
					green = ((cPixelArray[sourceIndex] >> 8) & 0xFF) * sourceAlpha / 255 +
						((canvas[destIndex] >> 8) & 0xFF) * destAlpha / 255;
					
					blue = ((cPixelArray[sourceIndex] >> 0) & 0xFF) * sourceAlpha / 255 +
						((canvas[destIndex] >> 0) & 0xFF) * destAlpha / 255;
					
					canvas[destIndex] = 0xFF000000 | (red << 16) | (green << 8) | (blue);
				}
				
				destIndex++;
				sourceIndex += deltaX;
			}
			
			destIndex += canvasWidth - plotWidth;
			sourceIndex -= plotWidth * deltaX;
			sourceIndex += deltaY;
		}
	}
	
	/**
	 * Accepts clipped dimensions only, does not support masks, but
	 * supports any mode. Faster than copyPixelsAlpha().
	 */
	protected void copyPixelsNoAlpha(int[] canvas, int canvasWidth, int sourceStart, int destStart,
			int plotWidth, int plotHeight, int deltaX, int deltaY) {
		
		int							yCounter;
		int							xCounter;
		int							sourceIndex;
		int							destIndex;
		
		sourceIndex = sourceStart;
		destIndex = destStart;
		
		for (yCounter = 0; yCounter < plotHeight; yCounter++) {
			for (xCounter = 0; xCounter < plotWidth; xCounter++) {
				canvas[destIndex] = cPixelArray[sourceIndex];
				
				destIndex++;
				sourceIndex += deltaX;
			}
			
			destIndex += canvasWidth - plotWidth;
			sourceIndex += deltaY - plotWidth * deltaX;
		}
	}
	
	/**
	 * Accepts clipped dimensions only, does not support masks, and
	 * only supports mode 0. Is generally faster than copyPixelsNoAlpha().
	 */
	protected void copyPixelsAC(int[] canvas, int canvasWidth, int sourceStart, int destStart,
			int plotWidth, int plotHeight, int deltaY) {
		
		int							yCounter;
		int							sourceIndex;
		int							destIndex;
		
		sourceIndex = sourceStart;
		destIndex = destStart;
		
		for (yCounter = 0; yCounter < plotHeight; yCounter++) {
			System.arraycopy(cPixelArray, sourceIndex, canvas, destIndex, plotWidth);
			
			destIndex += canvasWidth;
			sourceIndex += deltaY;
		}
	}
	
	/**
	 * Accepts clipped dimensions only, and determines which blitter
	 * to use. You can override it to provide additional blitters for
	 * special effects.
	 */
	protected void copyPixels(int[] canvas, int canvasWidth, int sourceStart, int destStart,
			int plotWidth, int plotHeight, int deltaX, int deltaY) {
		if (cAlphaMod != 255 || cHasMask)
			copyPixelsAlpha(canvas, canvasWidth, sourceStart, destStart, plotWidth, plotHeight, deltaX, deltaY);
		else if (deltaX == 1)
			copyPixelsAC(canvas, canvasWidth, sourceStart, destStart, plotWidth, plotHeight, deltaY);
		else
			copyPixelsNoAlpha(canvas, canvasWidth, sourceStart, destStart, plotWidth, plotHeight, deltaX, deltaY);
	}
	
	/**
	 * Takes screen-clipped coordinates and clips them further
	 * to make sure they stay within the buffer bounds.
	 */
	protected void drawVisible(int[] canvas, int canvasWidth,
			int localLeft, int localTop, int localRight, int localBottom,
			int leftInset, int topInset, int rightInset, int bottomInset,
			int deltaX, int deltaY, int mode) {
		// deltaX and deltaY refers to sourceIndex's modifications
		// destinationIndex will use 1 and canvasWidth respectively
		
		int							canvasHeight = canvas.length / canvasWidth;
		
		int							plotWidth;
		int							plotHeight;
		
		int							sourceStartX;
		int							sourceStartY;
		int							sourceStart;
		int							destStart;
		
		// clip
		if (localLeft < 0) {
			leftInset -= localLeft;
			localLeft = 0;
		}
		if (localTop < 0) {
			topInset -= localTop;
			localTop = 0;
		}
		if (localRight > canvasWidth) {
			rightInset += localRight - canvasWidth;
			localRight = canvasWidth;
		}
		if (localBottom > canvasHeight) {
			bottomInset += localBottom - canvasHeight;
			localBottom = canvasHeight;
		}
		
		plotWidth = localRight - localLeft;
		plotHeight = localBottom - localTop;
		
		// make sure size is still positive
		if (plotWidth <= 0 || plotHeight <= 0) {
			// clipped into non-existence
			return;
		}
		
		// calculate where source pixels index should start
		if (deltaX > 0)
			sourceStartX = leftInset;
		else
			sourceStartX = -(getWidth(mode) - leftInset - 1);
		
		if (deltaY > 0)
			sourceStartY = topInset;
		else
			sourceStartY = -(getHeight(mode) - topInset - 1);
		
		sourceStart = sourceStartX * deltaX + sourceStartY * deltaY;
		
		// destination pixel index always starts at top left
		destStart = localLeft + localTop * canvasWidth;
		
		copyPixels(canvas, canvasWidth, sourceStart, destStart,
				plotWidth, plotHeight, deltaX, deltaY);
	}
	
	/**
	 * Accepts any location, should be used with a wrapped buffer only.
	 * <ul>
	 * <li>globalLeft: location of image on map in pixels
	 * <li>globalTop: location of image on map in pixels
	 * <li>mode: rotated and/or mirrored
	 * <li>scrollX: the amount in pixels the map image has been scrolled left
	 * <li>scrollY: the amount in pixels the map image has been scrolled up
	 * </ul>
	 */
	public void drawWrapped(int[] canvas, int canvasWidth,
			int globalLeft, int globalTop, int mode,
			int scrollX, int scrollY) {
		
		int							canvasHeight = canvas.length / canvasWidth;
		
		// global... will not be clipped, level coordinates
		int							globalRight = globalLeft + getWidth(mode);
		int							globalBottom = globalTop + getHeight(mode);
		
		// local... will be clipped, canvas coordinates
		int							localLeft;
		int							localTop;
		int							localRight;
		int							localBottom;
		
		// inset is always >= 0
		int							leftInset = 0;
		int							topInset = 0;
		int							rightInset = 0;
		int							bottomInset = 0;
		
		int							deltaX;
		int							deltaY;
		
		// check if this image is visible
		if (globalRight <= scrollX || globalBottom <= scrollY ||
				globalLeft >= scrollX + canvasWidth || globalTop >= scrollY + canvasHeight)
			return;
		
		localLeft = globalLeft - canvasWidth * (scrollX / canvasWidth);
		localTop = globalTop - canvasHeight * (scrollY / canvasHeight);
		localRight = globalRight - canvasWidth * (scrollX / canvasWidth);
		localBottom = globalBottom - canvasHeight * (scrollY / canvasHeight);
		
		// trim image
		if (globalLeft < scrollX) {
			// falls off the left screen edge
			leftInset = scrollX - globalLeft;
			localLeft += leftInset;
		}
		if (globalTop < scrollY) {
			// falls off the top screen edge
			topInset = scrollY - globalTop;
			localTop += topInset;
		}
		if (globalRight > scrollX + canvasWidth) {
			// falls off the right screen edge
			rightInset = globalRight - (scrollX + canvasWidth);
			localRight -= rightInset;
		}
		if (globalBottom > scrollY + canvasHeight) {
			// falls off the bottom screen edge
			bottomInset = globalBottom - (scrollY + canvasHeight);
			localBottom -= bottomInset;
		}
		
		mode = mode % 8;
		
		if (mode == 0) {
			deltaX = 1;
			deltaY = cWidth;
		} else if (mode == 1) {
			deltaX = -cWidth;
			deltaY = 1;
		} else if (mode == 2) {
			deltaX = -1;
			deltaY = -cWidth;
		} else if (mode == 3) {
			deltaX = cWidth;
			deltaY = -1;
		} else if (mode == 4) {
			deltaX = -1;
			deltaY = cWidth;
		} else if (mode == 5) {
			deltaX = cWidth;
			deltaY = 1;
		} else if (mode == 6) {
			deltaX = 1;
			deltaY = -cWidth;
		} else if (mode == 7) {
			deltaX = -cWidth;
			deltaY = -1;
		} else
			return;
		
		try {
			if (localLeft < canvasWidth && localTop < canvasHeight) {
				// draw the bottom-right part (canvas coordinates)
				drawVisible(canvas, canvasWidth, localLeft, localTop, localRight, localBottom,
						leftInset, topInset, rightInset, bottomInset, deltaX, deltaY, mode);
			}
			
			if (localRight - canvasWidth > 0 && localBottom - canvasHeight > 0) {
				// draw the top-left part (canvas coordinates)
				drawVisible(canvas, canvasWidth, localLeft - canvasWidth, localTop - canvasHeight, localRight - canvasWidth, localBottom - canvasHeight,
						leftInset, topInset, rightInset, bottomInset, deltaX, deltaY, mode);
			}
			
			if (localRight - canvasWidth > 0 && localTop < canvasHeight) {
				// draw the bottom-left part (canvas coordinates)
				drawVisible(canvas, canvasWidth, localLeft - canvasWidth, localTop, localRight - canvasWidth, localBottom,
						leftInset, topInset, rightInset, bottomInset, deltaX, deltaY, mode);
			}
			
			if (localLeft < canvasWidth && localBottom - canvasHeight > 0) {
				// draw the top-right part (canvas coordinates)
				drawVisible(canvas, canvasWidth, localLeft, localTop - canvasHeight, localRight, localBottom - canvasHeight,
						leftInset, topInset, rightInset, bottomInset, deltaX, deltaY, mode);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// filter the exception through GOutput and provide extra information
			GOutput.logln("GPixels.drawWrapped() arguments: " + canvas.length + ", " + canvasWidth + "; " +
				globalLeft + ", " + globalTop + ", " + mode + "; " + scrollX + ", " + scrollY);
			GOutput.throwRuntime("GPixels.drawWrapped(): ArrayIndexOutOfBoundsException");
		}
	}
	
	
	/**
	 * Accepts any location, should be used with a non-wrapped buffer only.
	 */
	public void draw(int[] canvas, int canvasWidth, int left, int top, int mode) {
		int							canvasHeight = canvas.length / canvasWidth;
		
		int							deltaX;
		int							deltaY;
		
		mode = mode % 8;
		
		if (mode == 0) { deltaX = 1; deltaY = cWidth;
		} else if (mode == 1) { deltaX = -cWidth; deltaY = 1;
		} else if (mode == 2) { deltaX = -1; deltaY = -cWidth;
		} else if (mode == 3) { deltaX = cWidth; deltaY = -1;
		} else if (mode == 4) { deltaX = -1; deltaY = cWidth;
		} else if (mode == 5) { deltaX = cWidth; deltaY = 1;
		} else if (mode == 6) { deltaX = 1; deltaY = -cWidth;
		} else if (mode == 7) { deltaX = -cWidth; deltaY = -1;
		} else return;
		
		try {
			drawVisible(canvas, canvasWidth, left, top, left + getWidth(mode), top + getHeight(mode),
					0, 0, 0, 0, deltaX, deltaY, mode);
		} catch (ArrayIndexOutOfBoundsException e) {
			// filter the exception through GOutput and provide extra information
			GOutput.logln("GPixels.draw() arguments: " + canvas.length + ", " + canvasWidth + "; " +
				left + ", " + top + ", " + mode);
			GOutput.throwRuntime("GPixels.draw(): ArrayIndexOutOfBoundsException");
		}
	}
	
	/**
	 * Accepts clipped dimensions only.
	 * <ul>
	 * <li>sourceLeft: leftmost column copied
	 * <li>sourceTop: topmost row copied
	 * <li>sourceRight: rightmost column copied (-1)
	 * <li>sourceBottom: bottommost row copied (-1)
	 * <li>destLeft: tile's absolute left, whether it's visible or not
	 * <li>destTop: tile's absolute top, whether it's visible or not
	 * <p>
	 * <li>sourceStart: starting source pixel index
	 * <li>destStart: starting dest pixel index
	 * <li>width: displayed width
	 * <li>height: displayed height
	 * </ul>
	 */
	public void drawTile(int[] canvas, int canvasWidth, int sourceLeft, int sourceTop,
			int sourceRight, int sourceBottom, int destLeft, int destTop) {
		
		int							sourceStart;
		int							destStart;
		int							width;
		int							height;
		
		sourceStart = sourceLeft + sourceTop * cWidth;
		destStart = destLeft + sourceLeft + (destTop + sourceTop) * canvasWidth;
		width = sourceRight - sourceLeft;
		height = sourceBottom - sourceTop;
		
		if (width <= 0 || height <= 0)
			return;
		
		try {
			copyPixels(canvas, canvasWidth, sourceStart, destStart, width, height, 1, cWidth);
		} catch (ArrayIndexOutOfBoundsException e) {
			// filter the exception through GOutput and provide extra information
			GOutput.logln("GPixels.drawTile() arguments: " + canvas.length + ", " + canvasWidth + "; " +
				sourceLeft + ", " + sourceTop + ", " + sourceRight + ", " + sourceBottom + "; " +
				destLeft + ", " + destTop);
			GOutput.throwRuntime("GPixels.drawTile(): ArrayIndexOutOfBoundsException");
		}
	}
	
	/**
	 * Returns the mode that is flipped horizontally, compared to
	 * the passed mode.
	 */
	public static int getFlipHMode(int mode) {
		return (mode + 4) % 8;
	}
	
	/**
	 * Returns the mode that is flipped vertically, compared to
	 * the passed mode.
	 */
	public static int getFlipVMode(int mode) {
		if (mode < 4)
			return 4 + ((mode + 2) & 3);
		else
			return (mode + 2) & 3;
	}
	
	/**
	 * Returns the mode that is rotated the specified number of
	 * quarter-revolutions, compared to the passed mode.
	 */
	public static int getRotateMode(int mode, int rotate) {
		if (mode < 4)
			return (mode + rotate) & 3;
		else
			return 4 + ((mode + rotate) & 3);
	}
	
	public void copyTo(int[] canvas) {
		if (canvas != null && canvas.length == cPixelArray.length)
			System.arraycopy(cPixelArray, 0, canvas, 0, cPixelArray.length);
	}
	
	public int getAlphaMod() {
		return cAlphaMod;
	}
	
	/**
	 * Applies transparency to an image. Pass 255 to apply no extra
	 * transparency, or 0 for a completely transparent image.
	 * If the image is already transparent, the product of the
	 * opacity will be used.
	 */
	public void setAlphaMod(int alphaMod) {
		cAlphaMod = alphaMod;
		if (cAlphaMod < 0)
			cAlphaMod = 0;
		if (cAlphaMod > 255)
			cAlphaMod = 255;
	}
	
	/**
	 * Returns the width when drawn with mode 0.
	 */
	public int getWidth() {
		return cWidth;
	}
	
	/**
	 * Returns the width when drawn with mode 0.
	 */
	public int getHeight() {
		return cHeight;
	}
	
	/**
	 * Returns the width when drawn with the specified mode.
	 */
	public int getWidth(int mode) {
		if (mode % 2 == 0)
			return cWidth;
		return cHeight;
	}
	
	/**
	 * Returns the height when drawn with the specified mode.
	 */
	public int getHeight(int mode) {
		if (mode % 2 == 0)
			return cHeight;
		return cWidth;
	}
	
	public int getPixel(int pixelIndex) {
		if (pixelIndex < 0 || pixelIndex >= cPixelArray.length)
			GOutput.throwOB("pixelIndex", pixelIndex, 0, cPixelArray.length - 1);
		
		if (pixelIndex < 0)
			pixelIndex = 0;
		if (pixelIndex > cPixelArray.length - 1)
			pixelIndex = cPixelArray.length - 1;
		
		return cPixelArray[pixelIndex];
	}
	
	public int getPixel(int x, int y) {
		return getPixel(x + y * cWidth);
	}
	
	public void setPixel(int pixelIndex, int value) {
		if (pixelIndex < 0 || pixelIndex >= cPixelArray.length) {
			GOutput.throwOB("pixelIndex", pixelIndex, 0, cPixelArray.length - 1);
			return;
		}
		
		cPixelArray[pixelIndex] = value;
	}
	
	public void setPixel(int x, int y, int value) {
		setPixel(x + y * cWidth, value);
	}
}

