package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GProperties is used to hold the game-specific, saved
 * properties of GSpriteForms, GTileForms, GSprites
 * and GLevels. It can store numbers, flags and texts
 * (meaning unsigned shorts, booleans and Strings, but
 * avoiding name collisions).
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GProperties implements Cloneable
{
	public static final int			kMaxCount = 255;
	
	public static final int			kNumber = 0;
	public static final int			kFlag = 1;
	public static final int			kText = 2;
	
	protected int[]					cNumber;
	protected boolean[]				cFlag;
	protected String[]				cText;
	
	
	public GProperties() {
		clear();
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							numberCount;
		int							flagCount;
		int							textCount;
		
		numberCount = input.pullInt(8);
		flagCount = input.pullInt(8);
		textCount = input.pullInt(8);
		
		cNumber = new int[numberCount];
		for (int numberCounter = 0; numberCounter < numberCount; numberCounter++)
			cNumber[numberCounter] = input.pullInt(16);
		
		cFlag = new boolean[flagCount];
		for (int flagCounter = 0; flagCounter < flagCount; flagCounter++)
			cFlag[flagCounter] = input.pullBoolean();
		
		cText = new String[textCount];
		for (int textCounter = 0; textCounter < textCount; textCounter++)
			cText[textCounter] = input.pullString();
	}
	
	public Object clone() {
		GProperties					other = null;
		
		try { other = (GProperties) super.clone(); } catch (CloneNotSupportedException e) {}
		
		other.cNumber = new int[cNumber.length];
		System.arraycopy(cNumber, 0, other.cNumber, 0, cNumber.length);
		
		other.cFlag = new boolean[cFlag.length];
		System.arraycopy(cFlag, 0, other.cFlag, 0, cFlag.length);
		
		other.cText = new String[cText.length];
		System.arraycopy(cText, 0, other.cText, 0, cText.length);
		
		return other;
	}
	
	public void clear() {
		cNumber = new int[0];
		cFlag = new boolean[0];
		cText = new String[0];
	}
	
	/**
	 * Creates the specified number of data slots.
	 * A negative count will leave those properties untouched.
	 */
	public void clear(int numberCount, int flagCount, int textCount) {
		if (numberCount >= 0)
			cNumber = new int[numberCount];
		
		if (flagCount >= 0)
			cFlag = new boolean[flagCount];
		
		if (textCount >= 0) {
			cText = new String[textCount];
			
			for (int textCounter = 0; textCounter < textCount; textCounter++)
				cText[textCounter] = "";
		}
	}
	
	// Numbers:
	
	public int getNumberCount() {
		return cNumber.length;
	}
	
	public void setNumberCount(int numberCount) {
		int[]						oldNumber;
		
		if (numberCount < 0 || numberCount > kMaxCount) {
			GOutput.throwOB("numberCount", numberCount, 0, kMaxCount);
			return;
		}
		
		oldNumber = cNumber;
		cNumber = new int[numberCount];
		
		if (cNumber.length > oldNumber.length)
			System.arraycopy(oldNumber, 0, cNumber, 0, oldNumber.length);
		else
			System.arraycopy(oldNumber, 0, cNumber, 0, cNumber.length);
	}
	
	public int getNumber(int index) {
		if (index < 0 || index >= getNumberCount()) {
			GOutput.throwOB("index", index, 0, getNumberCount() - 1);
			return -1;
		}
		
		return cNumber[index];
	}
	
	public void setNumber(int index, int value) {
		if (index < 0 || index >= getNumberCount()) {
			GOutput.throwOB("index", index, 0, getNumberCount() - 1);
			return;
		}
		
		cNumber[index] = value;
	}
	
	// Flags:
	
	public int getFlagCount() {
		return cFlag.length;
	}
	
	public void setFlagCount(int flagCount) {
		boolean[]					oldFlag;
		
		if (flagCount < 0 || flagCount > kMaxCount) {
			GOutput.throwOB("flagCount", flagCount, 0, kMaxCount);
			return;
		}
		
		oldFlag = cFlag;
		cFlag = new boolean[flagCount];
		
		if (cFlag.length > oldFlag.length)
			System.arraycopy(oldFlag, 0, cFlag, 0, oldFlag.length);
		else
			System.arraycopy(oldFlag, 0, cFlag, 0, cFlag.length);
	}
	
	public boolean getFlag(int index) {
		if (index < 0 || index >= getFlagCount()) {
			GOutput.throwOB("index", index, 0, getFlagCount() - 1);
			return false;
		}
		
		return cFlag[index];
	}
	
	public void setFlag(int index, boolean value) {
		if (index < 0 || index >= getFlagCount()) {
			GOutput.throwOB("index", index, 0, getFlagCount() - 1);
			return;
		}
		
		cFlag[index] = value;
	}
	
	// Text:
	
	public int getTextCount() {
		return cText.length;
	}
	
	public void setTextCount(int textCount) {
		String[]					oldText;
		
		if (textCount < 0 || textCount > kMaxCount) {
			GOutput.throwOB("textCount", textCount, 0, kMaxCount);
			return;
		}
		
		oldText = cText;
		cText = new String[textCount];
		
		if (cText.length > oldText.length) {
			System.arraycopy(oldText, 0, cText, 0, oldText.length);
			for (int textCounter = oldText.length; textCounter < cText.length; textCounter++)
				cText[textCounter] = "";
		} else
			System.arraycopy(oldText, 0, cText, 0, cText.length);
	}
	
	public String getText(int index) {
		if (index < 0 || index >= getTextCount()) {
			GOutput.throwOB("index", index, 0, getTextCount() - 1);
			return null;
		}
		
		return cText[index];
	}
	
	public void setText(int index, String value) {
		if (index < 0 || index >= getTextCount()) {
			GOutput.throwOB("index", index, 0, getTextCount() - 1);
			return;
		}
		
		cText[index] = value;
	}
}

