package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GSprite is GameForm's sprite format. It holds data about the
 * current state of the sprite. A GSprite also has a reference to a
 * GSpriteForm, that will provide the data that is common to
 * all sprites of the samy type, such as the images.
 * <p>
 * Also, the GSprite can override certain default values, stored in
 * the GSpriteForm (cCanPassSolidTiles for instance). The
 * GSprite values are relative to the default values, so 0 or false
 * means "use the default". This means that if the default 
 * (GSpriteForm) value changes, the GSprite value will also
 * change. For example, if cCanPassSolidTiles is true, it means
 * that this GSprite's ability will be opposite to the other
 * GSprites of the same type. If cCollisionWidth is 2, this sprite
 * will be considered 2 pixels wider than the other GSprites of the
 * samy type. If you use the accessor methods, they will do the
 * conversion for you, but you will still need to keep in mind that
 * if you change the value in GSpriteForm, it will effectively
 * change in all GSprites too.
 * <p>
 * Only some of the GSprite values will be saved, others will be
 * set at run-time. If you still want to save them, you can of
 * course map them to values in GProperties.
 * <p>
 * All size values are in units of pixels. However, all location
 * and speed values are stored in units of 1/1024 pixels. The
 * sprite will still appear at a full-pixel location, but it can
 * have split-pixel speeds (i.e. less than one pixel per frame).
 * This also avoids most of the overhead when converting from
 * floating-point numbers to integers.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
*/


public class GSprite implements Cloneable
{
	// non-saved variables:
	protected GLevel				cLevel;
	
	// formIndex is saved in the GLevel, and
	// passed to the GSprite constructor
	protected int					cFormIndex;
	protected GSpriteForm			cForm;
	
	// saved variables:
	protected GProperties			cProperties;
	
	protected int					cLocationX;
	protected int					cLocationY;
	
	// relative to defaults, i.e. xor/delta
	protected boolean				cCanPassSolidTiles;
	protected int					cCollisionEffect; // can be used as a message to the other sprite
	protected boolean				cDetectCollision; // true if interacting with other sprites
	protected int					cCollisionWidth;
	protected int					cCollisionHeight;
	
	// non-saved variables:
	protected int					cSpeedX;
	protected int					cSpeedY;
	protected int					cCurrentImage;
	protected int					cCurrentImageMode;
	protected boolean				cTouchingSolid;
	
	
	public GSprite(int formIndex) {
		clear();
		cFormIndex = formIndex;
		cForm = GSpriteForm.get(cFormIndex);
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		if (input == null)
			throw new IOException("GSprite.readFromFile: invalid file");
		
		version = input.pullInt(8);
		if (version == 1) {
			cProperties.readFromFile(input);
			
			cLocationX = input.pullInt(16) << 10;
			cLocationY = input.pullInt(16) << 10;
			
			// due to relative values
			setCanPassSolidTiles(input.pullBoolean());
			setCollisionEffect(input.pullInt(16));
			setDetectCollision(input.pullBoolean());
			setCollisionWidth(input.pullInt(8));
			setCollisionHeight(input.pullInt(8));
			
			clearNonSaved();
		} else {
			throw new IOException("GSprite.readFromFile: Unknown version: " + version);
		}
	}
	
	public Object clone() {
		GSprite						other = null;
		
		try { other = (GSprite) super.clone(); } catch (CloneNotSupportedException e) {}
		
		other.cProperties = (GProperties) cProperties.clone();
		
		// Other will by design share cForm and cLevel with this.
		// When cloning a level, the level has to ensure that the new
		// GSprite is mapped to the correct level.
		
		return other;
	}
	
	public void clearNonSaved() {
		cSpeedX = 0;
		cSpeedY = 0;
		cCurrentImage = 0;
		cCurrentImageMode = 0;
		cTouchingSolid = false;
	}
	
	public void clear() {
		cFormIndex = 0;
		cForm = GSpriteForm.get(cFormIndex);
		cLocationX = 0;
		cLocationY = 0;
		
		cProperties = GFactory.createProperties();
		cProperties.clear(cForm.getInstanceNumberCount(), cForm.getInstanceFlagCount(), cForm.getInstanceTextCount());
		
		cCanPassSolidTiles = false; // default
		cCollisionEffect = 0; // default
		cDetectCollision = false; // default
		cCollisionWidth = 0; // default
		cCollisionHeight = 0; // default
		
		clearNonSaved();
	}
	
	/**
	 * Invoked by GLevel after runtime creation. If the sprite
	 * loaded from a file or cloned, initialize() will not be invoked.
	 */
	public void initialize(int x, int y) {
		cLocationX = x;
		cLocationY = y;
		
		cProperties.clear(cForm.getInstanceNumberCount(), 
				cForm.getInstanceFlagCount(), 
				cForm.getInstanceTextCount());
		
		// default relative values
		cCanPassSolidTiles = false;
		cCollisionEffect = 0;
		cDetectCollision = false;
		cCollisionWidth = 0;
		cCollisionHeight = 0;
		
		clearNonSaved();
	}
	
	/**
	 * Draws the sprite with the correct image on the specified wrapped
	 * pixel buffer (canvas). scrollX and scrollY describe how far the
	 * wrapped buffer has been scrolled.
	 */
	public void draw(int[] canvas, int canvasWidth, int scrollX, int scrollY) {
		GPixels						image;
		
		// check if sprite is visible
		if ((getRight() >> 10) < scrollX ||
				(getBottom() >> 10) < scrollY ||
				(getLeft() >> 10) >= scrollX + canvasWidth ||
				(getTop() >> 10) >= scrollY + canvas.length / canvasWidth)
			return;
		
		image = cForm.getImage(cCurrentImage);
		
		if (image != null)
			image.drawWrapped(canvas, canvasWidth, getLeft() >> 10, getTop() >> 10,
				cCurrentImageMode, scrollX, scrollY);
	}
	
	/**
	 * Moves the sprite that is marked as being able to pass solid
	 * tiles. cToucingSolid will be set if the sprite's collision
	 * bounds is covering one or more solid tiles.
	 */
	public void moveCanPassSolid() {
		int							x;
		int							y;
		
		setLocationX(getLocationX() + getSpeedX());
		setLocationY(getLocationY() + getSpeedY());
		
		setTouchingSolid(false);
		
		// detect touching solid
		for (y = (getCollisionTop() >> 10) / GTileForm.getTileHeight();
				y <= ((getCollisionBottom() >> 10) - 1) / GTileForm.getTileHeight();
				y++)
			if (y >= 0 && y < cLevel.getHeight())
				for (x = (getCollisionLeft() >> 10) / GTileForm.getTileWidth();
						x <= ((getCollisionRight() >> 10) - 1) / GTileForm.getTileWidth();
						x++)
					if (x >= 0 && x < cLevel.getWidth())
						if (GTileForm.get(cLevel.getTile(x, y)).getSolid())
							setTouchingSolid(true);
	}
	
	/**
	 * Move the sprite that is marked as not being able to pass
	 * solid tiles. Assumes that the sprite is not in a solid
	 * tile before moving.
	 * <p>
	 * A sprite incapable of passing solid tiles can not move
	 * faster that the tile size per frame, in either direction.
	 * <p>
	 * A non-square sprite incapable of passing solid tiles
	 * should not be rotated close to solid tiles, because if
	 * the new collision bounds might move a part of the sprite
	 * to a solid tile, causing unexpected behavior. If you do
	 * rotate a sprite close to solid tiles, make sure that the
	 * new collision bounds do not cover any solid tiles.
	 * <p>
	 * cToucingSolid will be set if the sprite was hindered in
	 * its path by a solid tile.
	 */
	public void moveCanNotPassSolid() {
		// assume not touching solid
		setTouchingSolid(false);
		
		if (getSpeedX() != 0) {
			// horizontal movement
			int						column;
			int						topRow;
			int						bottomRow;
			int						y;
			int						suggestedX;
			
			if (getSpeedX() > 0) {
				// moving right
				if (getSpeedX() > ((GTileForm.getTileWidth() - 1) << 10))
					setSpeedX((GTileForm.getTileWidth() - 1) << 10);
				
				column = ((getCollisionRight() + getSpeedX() - 1) >> 10) / GTileForm.getTileWidth();
				
				suggestedX = column * (GTileForm.getTileWidth() << 10) - (getCurrentCollisionWidth() << 9);
			} else {
				// moving left
				if (getSpeedX() < -((GTileForm.getTileWidth() - 1) << 10))
					setSpeedX(-((GTileForm.getTileWidth() - 1) << 10));
				
				column = ((getCollisionLeft() + getSpeedX()) >> 10) / GTileForm.getTileWidth();
				
				suggestedX = (column + 1) * (GTileForm.getTileWidth() << 10) + (getCurrentCollisionWidth() << 9);
			}
			
			topRow = (getCollisionTop() >> 10) / GTileForm.getTileHeight();
			bottomRow = ((getCollisionBottom() >> 10) - 1) / GTileForm.getTileHeight();
			
			if (column < cLevel.getWidth() && column >= 0) {
				if (topRow < 0) {
					// above level, trim search area
					topRow = 0;
				} else if (bottomRow >= cLevel.getHeight()) {
					// below level, trim search area
					bottomRow = cLevel.getHeight() - 1;
				}
				
				if (topRow <= bottomRow)
					for (y = topRow; y <= bottomRow; y++)
						if (column >= 0 && column < cLevel.getWidth() && y >= 0 && y < cLevel.getHeight())
							if (GTileForm.get(cLevel.getTile(column, y)).getSolid())
								if (getSpeedX() != 0) {
									// move sprite next to the solid tile, and stop it
									setLocationX(suggestedX);
									setSpeedX(0);
									setTouchingSolid(true);
								}
			}
			
			setLocationX(getLocationX() + getSpeedX());
		}
		
		if (getSpeedY() != 0) {
			// vertical movement
			int						row;
			int						leftColumn;
			int						rightColumn;
			int						x;
			int						suggestedY;
			
			if (getSpeedY() > 0) {
				// moving down
				if (getSpeedY() > ((GTileForm.getTileHeight() - 1) << 10))
					setSpeedY((GTileForm.getTileHeight() - 1) << 10);
				
				row = ((getCollisionBottom() + getSpeedY() - 1) >> 10) / GTileForm.getTileHeight();
				
				suggestedY = row * (GTileForm.getTileHeight() << 10) - (getCurrentCollisionHeight() << 9);
			} else {
				// moving up
				if (getSpeedY() < -((GTileForm.getTileHeight() - 1) << 10))
					setSpeedY(-((GTileForm.getTileHeight() - 1) << 10));
				
				row = ((getCollisionTop() + getSpeedY()) >> 10) / GTileForm.getTileHeight();
				
				suggestedY = (row + 1) * (GTileForm.getTileHeight() << 10) + (getCurrentCollisionHeight() << 9);
			}
			
			leftColumn = (getCollisionLeft() >> 10) / GTileForm.getTileWidth();
			rightColumn = ((getCollisionRight() >> 10) - 1) / GTileForm.getTileWidth();
			
			if (row < cLevel.getHeight() && row >= 0) {
				if (leftColumn < 0) {
					// left of level, trim search area
					leftColumn = 0;
				} else if (rightColumn >= cLevel.getWidth()) {
					// right of level, trim search area
					rightColumn = cLevel.getWidth() - 1;
				}
				
				if (leftColumn <= rightColumn)
					for (x = leftColumn; x <= rightColumn; x++)
						if (x >= 0 && x < cLevel.getWidth() && row >= 0 && row < cLevel.getHeight())
							if (GTileForm.get(cLevel.getTile(x, row)).getSolid())
								if (getSpeedY() != 0) {
									// move sprite next to the solid tile, and stop it
									setLocationY(suggestedY);
									setSpeedY(0);
									setTouchingSolid(true);
								}
			}
			
			setLocationY(getLocationY() + getSpeedY());
		}
	}
	
	public void move(int frame) {
		if (getCanPassSolidTiles())
			moveCanPassSolid();
		else
			moveCanNotPassSolid();
	}
	
	/**
	 * Override to customize sprite behavior. Invoked each frame
	 * before any sprite movement and collision detection.
	 * Because of the ordering of sprites in GLevel, you may
	 * create but not remove sprites in cLevel during preProcess().
	 */
	public void preProcess(int frame) {}
	
	/**
	 * Override to customize sprite behavior. Invoked each frame
	 * after all sprite movement and collision detection, before
	 * the screen updates. Because of the ordering of sprites in
	 * GLevel, you may destroy but not create sprites in cLevel
	 * during postProcess().
	 * <p>
	 * Because the GLevelCanvas may or may not be updated each
	 * frame, any lengthy operation in preparation for an update
	 * should, to provide the best performance, be moved to draw();
	 */
	public void postProcess(int frame) {}
	
	/**
	 * Override to customize sprite behavior. Because of the
	 * ordering of sprites in GLevel, you may not create or
	 * destroy sprites during collideWith().
	 * <p>
	 * If getDetectCollision() returns false, collideWith() will
	 * not be invoked.
	 */
	public void collideWith(GSprite other, int frame) {}
	
	public void updateNumberCount() { cProperties.setNumberCount(cForm.getInstanceNumberCount()); }
	public void updateFlagCount() { cProperties.setFlagCount(cForm.getInstanceFlagCount()); }
	public void updateTextCount() { cProperties.setTextCount(cForm.getInstanceTextCount()); }
	public void updateStatCount(int type) {
		if (type == GProperties.kNumber)
			updateNumberCount();
		else if (type == GProperties.kFlag)
			updateFlagCount();
		else if (type == GProperties.kText)
			updateTextCount();
	}
	
	public GLevel getLevel() { return cLevel; }
	
	public GProperties getProperties() { return cProperties; }
	public int getNumber(int index) { return cProperties.getNumber(index); }
	public boolean getFlag(int index) { return cProperties.getFlag(index); }
	public String getText(int index) { return cProperties.getText(index); }
	
	// relative values
	public boolean getCanPassSolidTiles() { return (cCanPassSolidTiles != cForm.getCanPassSolidTiles()); }
	public int getCollisionEffect() { return (cCollisionEffect + cForm.getCollisionEffect()); }
	public boolean getDetectCollision() { return (cDetectCollision != cForm.getDetectCollision()); }
	public int getCollisionWidth() { return (cCollisionWidth + cForm.getCollisionWidth()); }
	public int getCollisionHeight() { return (cCollisionHeight + cForm.getCollisionHeight()); }
	
	public int getFormIndex() { return cFormIndex; }
	public GSpriteForm getForm() { return cForm; }
	public int getNativeWidth() { return cForm.getImage(0).getWidth(); }
	public int getNativeHeight() { return cForm.getImage(0).getHeight(); }
	
	public int getCurrentWidth() {
		if (cCurrentImageMode % 2 == 0) // rotated even number of times
			return cForm.getImage(cCurrentImage).getWidth();
		else // rotated odd number of times
			return cForm.getImage(cCurrentImage).getHeight();
	}
	
	public int getCurrentHeight() {
		if (cCurrentImageMode % 2 == 0) // rotated even number of times
			return cForm.getImage(cCurrentImage).getHeight();
		else // rotated odd number of times
			return cForm.getImage(cCurrentImage).getWidth();
	}
	
	public int getCurrentCollisionWidth() {
		if (cCurrentImageMode % 2 == 0) // rotated even number of times
			return getCollisionWidth();
		else // rotated odd number of times
			return getCollisionHeight();
	}
	
	public int getCurrentCollisionHeight() {
		if (cCurrentImageMode % 2 == 0) // rotated even number of times
			return getCollisionHeight();
		else // rotated odd number of times
			return getCollisionWidth();
	}
	
	public int getLocationX() { return cLocationX; }
	public int getLocationY() { return cLocationY; }
	public int getLeft() { return cLocationX - (getCurrentWidth() << 9); }
	public int getTop() { return cLocationY - (getCurrentHeight() << 9); }
	public int getRight() { return cLocationX + (getCurrentWidth() << 9); }
	public int getBottom() { return cLocationY + (getCurrentHeight() << 9); }
	public int getCollisionLeft() { return cLocationX - (getCurrentCollisionWidth() << 9); }
	public int getCollisionTop() { return cLocationY - (getCurrentCollisionHeight() << 9); }
	public int getCollisionRight() { return cLocationX + (getCurrentCollisionWidth() << 9); }
	public int getCollisionBottom() { return cLocationY + (getCurrentCollisionHeight() << 9); }
	public int getTileX() { return ((cLocationX >> 10) / GTileForm.getTileWidth()); }
	public int getTileY() { return ((cLocationY >> 10) / GTileForm.getTileHeight()); }
	public int getSpeedX() { return cSpeedX; }
	public int getSpeedY() { return cSpeedY; }
	public int getCurrentImage() { return cCurrentImage; }
	public int getCurrentImageMode() { return cCurrentImageMode; }
	public boolean getTouchingSolid() { return cTouchingSolid; }
	
	public void setLevel(GLevel level) { cLevel = level; }
	
	// relative values
	public void setCanPassSolidTiles(boolean canPassSolidTiles) { cCanPassSolidTiles = cForm.getCanPassSolidTiles() != canPassSolidTiles; }
	public void setCollisionEffect(int collisionEffect) { cCollisionEffect = collisionEffect - cForm.getCollisionEffect(); }
	public void setDetectCollision(boolean detectCollision) { cDetectCollision = cForm.getDetectCollision() != detectCollision; }
	public void setCollisionWidth(int collisionWidth) { cCollisionWidth = collisionWidth - cForm.getCollisionWidth(); }
	public void setCollisionHeight(int collisionHeight) { cCollisionHeight = collisionHeight - cForm.getCollisionHeight(); }
	
	public void setFormIndex(int index) {
		if (index < 0 || index > GSpriteForm.getCount()) {
			GOutput.throwOB("index", index, 0, GSpriteForm.getCount() - 1);
			return;
		}
		cFormIndex = index;
		cForm = GSpriteForm.get(cFormIndex);
	}
	
	public void setLocationX(int locationX) { cLocationX = locationX; }
	public void setLocationY(int locationY) { cLocationY = locationY; }
	public void setLocation(int locationX, int locationY) { cLocationX = locationX; cLocationY = locationY; }
	public void setSpeedX(int speedX) { cSpeedX = speedX; }
	public void setSpeedY(int speedY) { cSpeedY = speedY; }
	public void setSpeed(int speedX, int speedY) { cSpeedX = speedX; cSpeedY = speedY; }
	public void setCurrentImage(int currentImage) {
		if (currentImage < 0 || currentImage > cForm.getImageCount()) {
			GOutput.throwOB("currentImage", currentImage, 0, cForm.getImageCount() - 1);
			return;
		}
		cCurrentImage = currentImage;
	}
	
	/**
	 * Changes the image mode used when drawing this GSprite.
	 * A non-square sprite incapable of passing solid tiles
	 * should not be rotated close to solid tiles, because if
	 * the new collision bounds might move a part of the sprite
	 * to a solid tile, causing unexpected behavior in the
	 * moveCanNotPassSolid() method. If you do rotate a sprite
	 * close to solid tiles, make sure that the new collision
	 * bounds do not cover any solid tiles.
	 * <p>
	 * See GPixels for information about image modes.
	 */
	public void setCurrentImageMode(int currentImageMode) { cCurrentImageMode = currentImageMode; }
	public void setTouchingSolid(boolean touchingSolid) { cTouchingSolid = touchingSolid; }
}

