package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GSpriteForm is GameForm's sprite form format. It holds
 * the data that is shared between all the sprites of a certain
 * type, such as the images and certain other properties. An actual
 * GSprite in a level will keep a reference to a GSpriteForm,
 * and use the GSpriteForm data where appropriate.
 * <p>
 * Additionally, the GSpriteForm holds the default values for
 * some of the GSprite properties (cCanPassSolidTiles for
 * instance). You can easily change the behavior of all sprites of
 * a certain type by changing the GSpriteForm value, or change
 * the behavior of a single sprite by changing the GSprite value.
 * <p>
 * Most of the time, you will not interact directly with the
 * GSpriteForm, instead using the GSprite objects.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GSpriteForm implements Cloneable
{
	protected static GSpriteForm[]	sSpriteForm;
	
	protected static int			sNumberCount; // all sprite forms' GProperties have the same number of numbers
	protected static int			sFlagCount; // all sprite forms' GProperties have the same number of flags
	protected static int			sTextCount; // all sprite forms' GProperties have the same number of texts
	
	// saved variables
	protected GProperties			cProperties;
	protected GPixels[]				cImage;
	
	// defaults for instances
	protected boolean				cCanPassSolidTiles;
	protected int					cCollisionEffect;
	protected boolean				cDetectCollision;
	protected int					cCollisionWidth;
	protected int					cCollisionHeight;
	
	// instance-static, form-specific
	protected int					cInstanceNumberCount; // number of numbers in GSprites
	protected int					cInstanceFlagCount; // number of flags in GSprites
	protected int					cInstanceTextCount; // number of texts in GSprites
	
	
	public GSpriteForm() {
		clear();
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		if (input == null)
			throw new IOException("GSpriteForm.readFromFile: invalid file");
		
		version = input.pullInt(8);
		if (version == 1) {
			int						imageCounter;
			int						stateCounter;
			
			cProperties.readFromFile(input);
			
			cImage = new GPixels[input.pullInt(8)];
			for (imageCounter = 0; imageCounter < cImage.length; imageCounter++)
				cImage[imageCounter] = GFactory.createPixels(input);
			
			cCanPassSolidTiles = input.pullBoolean();
			cCollisionEffect = input.pullInt(16);
			cDetectCollision = input.pullBoolean();
			cCollisionWidth = input.pullInt(8);
			cCollisionHeight = input.pullInt(8);
			
			cInstanceNumberCount = input.pullInt(8);
			cInstanceFlagCount = input.pullInt(8);
			cInstanceTextCount = input.pullInt(8);
		} else {
			throw new IOException("GSpriteForm.readFromFile: Unknown version: " + version);
		}
	}
	
	public Object clone() {
		GSpriteForm					other = null;
		
		try { other = (GSpriteForm) super.clone(); } catch (CloneNotSupportedException e) {}
		
		// deep clone of properties
		other.cProperties = (GProperties) cProperties.clone();
		
		// deep clone of images
		other.cImage = new GPixels[cImage.length];
		for (int index = 0; index < cImage.length; index++)
			other.cImage[index] = (GPixels) cImage[index].clone();
		
		return other;
	}
	
	public void clear() {
		cProperties = GFactory.createProperties();
		cProperties.clear(sNumberCount, sFlagCount, sTextCount);
		
		cImage = new GPixels[1];
		cImage[0] = GFactory.createPixels();
	}
	
	public int getImageCount() {
		if (cImage == null)
			return 0;
		else
			return cImage.length;
	}
	
	public GPixels getImage(int index) {
		if (index < 0 || index >= getImageCount())
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
		
		if (index < 0)
			index = 0;
		if (index > getImageCount() - 1)
			index = getImageCount() - 1;
		
		return cImage[index];
	}
	
	// note: use the functions of GSprite if appropriate
	public GProperties getProperties() { return cProperties; }
	public int getNumber(int index) { return cProperties.getNumber(index); }
	public boolean getFlag(int index) { return cProperties.getFlag(index); }
	public String getText(int index) { return cProperties.getText(index); }
	
	public boolean getCanPassSolidTiles() { return cCanPassSolidTiles; }
	public int getCollisionEffect() { return cCollisionEffect; }
	public boolean getDetectCollision() { return cDetectCollision; }
	public int getCollisionWidth() { return cCollisionWidth; }
	public int getCollisionHeight() { return cCollisionHeight; }
	
	public void setCanPassSolidTiles(boolean canPassSolidTiles) { cCanPassSolidTiles = canPassSolidTiles; }
	public void setCollisionEffect(int collisionEffect) { cCollisionEffect = collisionEffect; }
	public void setDetectCollision(boolean detectCollision) { cDetectCollision = detectCollision; }
	public void setCollisionWidth(int collisionWidth) { cCollisionWidth = collisionWidth; }
	public void setCollisionHeight(int collisionHeight) { cCollisionHeight = collisionHeight; }
	
	public int getInstanceNumberCount() {
		return cInstanceNumberCount;
	}
	
	public void setInstanceNumberCount(int count) {
		if (count < 0 || count > GProperties.kMaxCount) {
			GOutput.throwOB("count", count, 0, GProperties.kMaxCount);
			return;
		}
		
		cInstanceNumberCount = count;
		
		GSpriteForm.updateSpriteInstanceStatCount(GProperties.kNumber);
	}
	
	public int getInstanceFlagCount() {
		return cInstanceFlagCount;
	}
	
	public void setInstanceFlagCount(int count) {
		if (count < 0 || count > GProperties.kMaxCount) {
			GOutput.throwOB("count", count, 0, GProperties.kMaxCount);
			return;
		}
		
		cInstanceFlagCount = count;
		
		GSpriteForm.updateSpriteInstanceStatCount(GProperties.kFlag);
	}
	
	public int getInstanceTextCount() {
		return cInstanceTextCount;
	}
	
	public void setInstanceTextCount(int count) {
		if (count < 0 || count > GProperties.kMaxCount) {
			GOutput.throwOB("count", count, 0, GProperties.kMaxCount);
			return;
		}
		
		cInstanceTextCount = count;
		
		GSpriteForm.updateSpriteInstanceStatCount(GProperties.kText);
	}
	
	public static int getStaticNumberCount() { return sNumberCount; }
	
	public static void setStaticNumberCount(int numberCount) {
		int							spriteFormCounter;
		
		if (numberCount < 0 || numberCount > GProperties.kMaxCount) {
			GOutput.throwOB("numberCount", numberCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sNumberCount = numberCount;
		
		for (spriteFormCounter = 0; spriteFormCounter < GSpriteForm.getCount(); spriteFormCounter++)
			sSpriteForm[spriteFormCounter].getProperties().setNumberCount(sNumberCount);
	}
	
	public static int getStaticFlagCount() { return sFlagCount; }
	
	public static void setStaticFlagCount(int flagCount) {
		int							spriteFormCounter;
		
		if (flagCount < 0 || flagCount > GProperties.kMaxCount) {
			GOutput.throwOB("flagCount", flagCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sFlagCount = flagCount;
		
		for (spriteFormCounter = 0; spriteFormCounter < GSpriteForm.getCount(); spriteFormCounter++)
			sSpriteForm[spriteFormCounter].getProperties().setFlagCount(sFlagCount);
	}
	
	public static int getStaticTextCount() { return sTextCount; }
	
	public static void setStaticTextCount(int textCount) {
		int							spriteFormCounter;
		
		if (textCount < 0 || textCount > GProperties.kMaxCount) {
			GOutput.throwOB("textCount", textCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sTextCount = textCount;
		
		for (spriteFormCounter = 0; spriteFormCounter < GSpriteForm.getCount(); spriteFormCounter++)
			sSpriteForm[spriteFormCounter].getProperties().setTextCount(sTextCount);
	}
	
	public static void readAllFromFile(GBitReader input) throws IOException {
		int							version;
		int							spriteCount;
		int							spriteCounter;
		
		GOutput.logln("reading sprite forms");
		
		if (input == null)
			throw new IOException("GSpriteForm.readAllFromFile: invalid file");
		
		version = input.pullInt(8);
		
		GOutput.logln("version " + version);
		
		if (version == 1) {
			GOutput.log("id ");
			
			sNumberCount = input.pullInt(8);
			sFlagCount = input.pullInt(8);
			sTextCount = input.pullInt(8);
			
			spriteCount = input.pullInt(8);
			if (spriteCount == 0)
				throw new IOException("GSpriteForm.readAllFromFile: invalid file format");
			
			sSpriteForm = new GSpriteForm[spriteCount];
			for (spriteCounter = 0; spriteCounter < spriteCount; spriteCounter++) {
				GOutput.log(spriteCounter + " ");
				sSpriteForm[spriteCounter] = GFactory.createSpriteForm(input);
			}
			GOutput.logln("done");
		} else {
			throw new IOException("GSpriteForm.readAllFromFile: unknown version: " + version);
		}
	}
	
	public static void clearAll() {
		sSpriteForm = new GSpriteForm[1];
		sSpriteForm[0] = GFactory.createSpriteForm();
	}
	
	public static int getCount() {
		if (sSpriteForm == null)
			return 0;
		else
			return sSpriteForm.length;
	}
	
	public static GSpriteForm get(int index) {
		if (index < 0 || index >= getCount())
			GOutput.throwOB("index", index, 0, getCount() - 1);
		
		if (index < 0)
			index = 0;
		if (index > getCount() - 1)
			index = getCount() - 1;
		
		return sSpriteForm[index];
	}
	
	public static void updateSpriteInstanceStatCount(int type) {
		int							levelCounter;
		
		// update all sprites in all levels
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GLevel.get(levelCounter).updateSpriteInstanceStatCount(type);
	}
}

