package glib;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/**
 * GTileForm is GameForm's tile format. It holds the data that
 * is shared between all the tiles of a certain type (in fact, there
 * is generally no other data). An actual tile in a level will be
 * stored as an index to a GTileForm.
 *
 * @version 1.0 08/21/2001
 * @author Bjorn Carlin
 */


public class GTileForm implements Cloneable
{
	protected static GTileForm[]	sTileForm;
	
	protected static int			sNumberCount;
	protected static int			sFlagCount;
	protected static int			sTextCount;
	
	// saved variables:
	protected GProperties			cProperties;
	
	protected GPixels[]				cImage;
	protected int					cAnimationPeriod;
	protected boolean				cSolid;
	
	
	public GTileForm() {
		clear();
	}
	
	public void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		if (input == null)
			throw new IOException("GTileForm.readFromFile: invalid file");
		
		version = input.pullInt(8);
		
		if (version == 1) {
			int						imageCounter;
			
			cProperties.readFromFile(input);
			
			cImage = new GPixels[input.pullInt(8)];
			for (imageCounter = 0; imageCounter < cImage.length; imageCounter++)
				cImage[imageCounter] = GFactory.createPixels(input);
			
			cAnimationPeriod = input.pullInt(8);
			cSolid = input.pullBoolean();
		} else {
			throw new IOException("GTileForm.readFromFile: Unknown version: " + version);
		}
	}
	
	public Object clone() {
		GTileForm					other = null;
		
		try { other = (GTileForm) super.clone(); } catch (CloneNotSupportedException e) {}
		
		// deep clone of properties
		other.cProperties = (GProperties) cProperties.clone();
		
		// deep clone of images
		other.cImage = new GPixels[cImage.length];
		for (int index = 0; index < cImage.length; index++)
			other.cImage[index] = (GPixels) cImage[index].clone();
		
		return other;
	}
	
	public void clear() {
		cProperties = GFactory.createProperties();
		cProperties.clear(sNumberCount, sFlagCount, sTextCount);
		
		cImage = new GPixels[1];
		cImage[0] = GFactory.createPixels();
		
		cAnimationPeriod = 0;
		cSolid = false;
	}
	
	public void draw(int frame, int[] canvas, int canvasWidth, int sourceLeft, int sourceTop,
			int sourceRight, int sourceBottom, int destLeft, int destTop) {
		getFrameImage(frame).drawTile(canvas, canvasWidth, sourceLeft, sourceTop,
				sourceRight, sourceBottom, destLeft, destTop);
	}
	
	public GProperties getProperties() { return cProperties; }
	public int getNumber(int index) { return cProperties.getNumber(index); }
	public boolean getFlag(int index) { return cProperties.getFlag(index); }
	public String getText(int index) { return cProperties.getText(index); }
	
	public int getImageCount() {
		return cImage.length;
	}
	
	/**
	 * Returns the image that should be displayed on the specified frame.
	 */
	public GPixels getFrameImage(int frame) {
		if (cAnimationPeriod < 1 || getImageCount() == 1 || frame < 0)
			return cImage[0];
		
		return cImage[(frame / cAnimationPeriod) % getImageCount()];
	}
	
	public GPixels getImage(int index) {
		if (index < 0 || index >= getImageCount())
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
		
		if (index < 0)
			index = 0;
		if (index > getImageCount() - 1)
			index = getImageCount() - 1;
		
		return cImage[index];
	}
	
	public int getAnimationPeriod() {
		return cAnimationPeriod;
	}
	
	public void setAnimationPeriod(int animationPeriod) {
		cAnimationPeriod = animationPeriod;
	}
	
	public boolean getSolid() {
		return cSolid;
	}
	
	public void setSolid(boolean solid) {
		cSolid = solid;
	}
	
	public static int getStaticNumberCount() { return sNumberCount; }
	
	public static void setStaticNumberCount(int numberCount) {
		int							tileFormCounter;
		
		if (numberCount < 0 || numberCount > GProperties.kMaxCount) {
			GOutput.throwOB("numberCount", numberCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sNumberCount = numberCount;
		
		for (tileFormCounter = 0; tileFormCounter < getCount(); tileFormCounter++)
			sTileForm[tileFormCounter].getProperties().setNumberCount(sNumberCount);
	}
	
	public static int getStaticFlagCount() { return sFlagCount; }
	
	public static void setStaticFlagCount(int flagCount) {
		int							tileFormCounter;
		
		if (flagCount < 0 || flagCount > GProperties.kMaxCount) {
			GOutput.throwOB("flagCount", flagCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sFlagCount = flagCount;
		
		for (tileFormCounter = 0; tileFormCounter < getCount(); tileFormCounter++)
			sTileForm[tileFormCounter].getProperties().setFlagCount(sFlagCount);
	}
	
	public static int getStaticTextCount() { return sTextCount; }
	
	public static void setStaticTextCount(int textCount) {
		int							tileFormCounter;
		
		if (textCount < 0 || textCount > GProperties.kMaxCount) {
			GOutput.throwOB("textCount", textCount, 0, GProperties.kMaxCount);
			return;
		}
		
		sTextCount = textCount;
		
		for (tileFormCounter = 0; tileFormCounter < getCount(); tileFormCounter++)
			sTileForm[tileFormCounter].getProperties().setTextCount(sTextCount);
	}
	
	public static void readAllFromFile(GBitReader input) throws IOException {
		int							version;
		int							tileCount;
		int							tileCounter;
		
		GOutput.logln("reading tile forms");
		
		if (input == null)
			throw new IOException("GTileForm.readAllFromFile: invalid file");
		
		version = input.pullInt(8);
		
		GOutput.logln("version " + version);
		
		if (version == 1) {
			GOutput.log("id ");
			
			sNumberCount = input.pullInt(8);
			sFlagCount = input.pullInt(8);
			sTextCount = input.pullInt(8);
			
			tileCount = input.pullInt(8);
			if (tileCount == 0)
				throw new IOException("GTileForm.readAllFromFile: invalid file format");
			
			sTileForm = new GTileForm[tileCount];
			for (tileCounter = 0; tileCounter < tileCount; tileCounter++) {
				GOutput.log(tileCounter + " ");
				sTileForm[tileCounter] = GFactory.createTileForm(input);
			}
			GOutput.logln("done");
		} else {
			throw new IOException("GTileForm.readAllFromFile: unknown version: " + version);
		}
	}
	
	public static void clearAll() {
		sTileForm = new GTileForm[1];
		sTileForm[0] = GFactory.createTileForm();
	}
	
	public static int getTileWidth() {
		if (sTileForm == null ||
				sTileForm[0] == null ||
				sTileForm[0].getImage(0) == null)
			return 16;
		else
			return sTileForm[0].getImage(0).getWidth();
	}
	
	public static int getTileHeight() {
		if (sTileForm == null ||
				sTileForm[0] == null ||
				sTileForm[0].getImage(0) == null)
			return 16;
		else
			return sTileForm[0].getImage(0).getHeight();
	}
	
	public static int getCount() {
		if (sTileForm == null)
			return 0;
		else
			return sTileForm.length;
	}
	
	public static GTileForm get(int index) {
		if (index < 0 || index >= getCount())
			GOutput.throwOB("index", index, 0, getCount() - 1);
		
		if (index < 0)
			index = 0;
		if (index > getCount() - 1)
			index = getCount() - 1;
		
		return sTileForm[index];
	}
}

