/***************************************************************

Bouncing is an example of how you can extend the graphics
engine. It subclasses GLevelCanvas and GTileForm to allow tiles
to be drawn both below and above the sprites (the darker
diamonds in the center, and the black, covered, tiles), and to
be animated by drawing two different images with different
transparency (the flashing walls).

Bouncing also subclasses GSprite and GPixels to change the color
of a single (gray) ball image. (Wrapping uses a different
technique to accomplish a similar effect, using two images drawn
with different levels of transparency).

Bouncing also shows how to "prime" the Just-in-Time (JIT)
compiler. The JIT compiler caches interpreted bytecode the first
time it is run, so it can be run a lot faster the second time
around. After loading the data file, Bouncing runs a couple of
frames of a "fake game" in the background, giving the JIT
compiler a chance to compile the bytecode ahead of time. This
means the game takes a little time to start up, but runs faster.
Toggle kPrimeJIT to see the difference in missed frames.

Bouncing uses the sprites' GProperties objects to store the color
and speed information.


The source for BouncingApplet is in the public domain. This
means you can use it in your own projects, whole or in part.
Please note that GameForm is not in the public domain, see the
GameForm source, documentation, or the GameForm homepage for
information about the allowed and disallowed usage of GameForm.


GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;
import java.net.*;


public class BouncingApplet extends GameApplet implements GLoadListener, MouseListener
{
	// keys
	static final int				kLeft = 0;
	static final int				kRight = 1;
	static final int				kUp = 2;
	static final int				kDown = 3;
	
	static final boolean			kPrimeJIT = true; // see newProgress for details
	
	
	BouncingLevelCanvas				cCanvas;
	
	// a "dirty" working copy of the current level
	GLevel							cLevel;
	
	// the BouncingPlayer will set these variables, so this BouncingApplet
	// will know where to scroll to center on the player
	int								cTargetCenterX;
	int								cTargetCenterY;
	
	// file loading status
	boolean							cDoneLoading;
	boolean							cFailedLoading;
	
	// current game information
	int								cPoints;
	int								cUpdateFrame;
	
	
	public void init() {
		// prepare GameApplet
		super.init(kNoBuffer);
		cFrameLength = 40; // 40 milliseconds, 25 frames per second
		setKeys(5, 3,
				new int[] {KeyEvent.VK_LEFT,
							KeyEvent.VK_RIGHT,
							KeyEvent.VK_UP,
							KeyEvent.VK_DOWN},
				new int[] {kKeyContinous,
							kKeyContinous,
							kKeyContinous,
							kKeyContinous});
		
		// Toggle BouncingLevelCanvas debugging when typing "debug"
		setCheats(new int[][] {{
				KeyEvent.VK_D,
				KeyEvent.VK_E,
				KeyEvent.VK_B,
				KeyEvent.VK_U,
				KeyEvent.VK_G}});
		
		// prepare GameForm
		GFactory.setFactory(new BouncingFactory());
		GOutput.sMute = true;
		
		BouncingPlayer.cApplet = this;
		
		// init buffer
		cCanvas = new BouncingLevelCanvas(this, null, getSize().width, getSize().height);
		
		// share cCanvas' buffer with GameApplet
		cBuffer = cCanvas.getBuffer();
		cBufferGraphics = cBuffer.getGraphics();
		
		// start loading data
		cDoneLoading = false;
		cFailedLoading = false;
		try {
			GFileManager.load(new URL(getCodeBase(), getParameter("file")), this);
		} catch (MalformedURLException e) {
			cFailedLoading = true;
		}
		addMouseListener(this);
	}
	
	public void pauseGame(boolean pause) {
		super.pauseGame(pause);
	}
	
	public void newGame() {
		if (!cDoneLoading)
			return;
		
		cPoints = 0;
		cUpdateFrame = 0;
		
		// create a working copy (clone) of the first level so level 0
		// can be used several times without reloading it from file
		cLevel = GFactory.createLevel(GLevel.get(0));
		cCanvas.setLevel(cLevel);
		cBuffer = cCanvas.getBuffer();
		cBufferGraphics = cBuffer.getGraphics();
		
		// start the flashing bricks
		BouncingTile.cAnimate = true;
		
		// because super.newGame start the timer, it is
		// generally called after other newGame code
		super.newGame();
	}
	
	public void endGame() {
		super.endGame();
	}
	
	public void cheatActivated(int cheatID) {
		if (cheatID == 0) { // debug
			cCanvas.cDebug = !cCanvas.cDebug;
			if (!cCanvas.cDebug)
				cCanvas.setNeedUpdate(true);
		}
	}
	
	/*
	Move the game state to the next (i.e. current) frame.
	advanceFrame() is synchronized to keep the game thread from changing
	the data while the event thread is updating the screen.
	*/
	public synchronized void advanceFrame() {
		int							targetScrollX;
		int							targetScrollY;
		
		// let GameForm do its thing
		cLevel.nextFrame(cFrame);
		
		// update scrolling, centering on the player with a slight delay
		targetScrollX = cTargetCenterX - getSize().width / 2;
		targetScrollY = cTargetCenterY - getSize().height / 2;
		
		cCanvas.setScrollX((cCanvas.getScrollX() * 5 + targetScrollX) / 6);
		cCanvas.setScrollY((cCanvas.getScrollY() * 5 + targetScrollY) / 6);
	}
	
	public void newProgress(int bitsRead) {
		if (bitsRead == GFileManager.kFinished) {
			if (kPrimeJIT) {
				/* Run the code below to prime the JIT (just-in-time)
				compiler. Notice that less frames are missed in the beginning
				of the first game if you prime the JIT compiler, because
				bytecode runs much slower the first time. Notice also that
				less frames are generally missed in the second game than the
				first, and by running the code below most of the code will be
				run once or twice, before any game starts. If you have a fast
				computer that's not missing any frames, try changing the
				cFrameLength variable in init() to a lower value, to increase
				the frame rate.
				*/
				GLevel				level;
				
				level = GFactory.createLevel(GLevel.get(0));
				cCanvas.setLevel(level);
				
				level.nextFrame(1);
				cCanvas.paint(null, 1);
				level.nextFrame(2);
				cCanvas.paint(null, 2);
			}
			cDoneLoading = true;
			repaint();
		} else if (bitsRead == GFileManager.kFailed) {
			cFailedLoading = true;
			repaint();
		} else {
			// If the data file is large, this would be a good place
			// to update a progress bar.
		}
	}
	
	/*
	repaint() has been called.
	update() is synchronized to keep the game thread from changing
	the data while this (event) thread is updating the screen.
	*/
	public synchronized void update(Graphics theGraphics) {
		// paint to the buffer
		if (cGameInProgress) {
			// a game is running, paint the level to cBuffer
			cCanvas.paint(null, cFrame);
			
			// add the score and time floating over level
			cBufferGraphics.setFont(new Font("SansSerif", Font.PLAIN, 10));
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.drawString(cPoints + " points", 4, 16);
			cBufferGraphics.drawString(cFrame * cFrameLength / 1000 + " s", 4, 32);
			
			cBufferGraphics.setColor(Color.white);
			cBufferGraphics.drawString(cPoints + " points", 3, 15);
			cBufferGraphics.drawString(cFrame * cFrameLength / 1000 + " s", 3, 31);
			
			// keep track of how often we update, and display statistics
			cUpdateFrame++;
			if (cFrame > 0) {
				cBufferGraphics.setColor(Color.black);
				cBufferGraphics.drawString(cFrame - cUpdateFrame + " (" + (cFrame - cUpdateFrame) * 100 / cFrame + "%) updates missed", getSize().width / 2 + 4, 16);
				
				cBufferGraphics.setColor(Color.white);
				cBufferGraphics.drawString(cFrame - cUpdateFrame + " (" + (cFrame - cUpdateFrame) * 100 / cFrame + "%) updates missed", getSize().width / 2 + 3, 15);
			}
		} else {
			// no game is running, display the title screen
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.fillRect(0, 0, getSize().width, getSize().height);
			
			cBufferGraphics.setFont(new Font("SansSerif", Font.PLAIN, 14));
			cBufferGraphics.setColor(Color.white);
			if (cFailedLoading)
				cBufferGraphics.drawString("Failed Loading Data", (getSize().width - cBufferGraphics.getFontMetrics().stringWidth("Failed Loading Data")) / 2, getSize().height / 2 + 10);
			else if (cDoneLoading)
				cBufferGraphics.drawString("Click here to Start", (getSize().width - cBufferGraphics.getFontMetrics().stringWidth("Press Space to Start")) / 2, getSize().height / 2 + 10);
			else
				cBufferGraphics.drawString("Loading Data... Please Wait.", (getSize().width - cBufferGraphics.getFontMetrics().stringWidth("Loading Data... Please Wait.")) / 2, getSize().height / 2 + 10);
		}
		
		// copy buffer to screen
		theGraphics.drawImage(cBuffer, 0, 0, null);
	}
	
	/*
	The view is "damaged"
	*/
	public void paint(Graphics theGraphics) {
		if (cGameInProgress) {
			// because a game has been started, the buffer should be up-to-date
			// copy buffer to screen
			theGraphics.drawImage(cBuffer, 0, 0, null);
		} else {
			// the buffer may or may not be up-to-date
			// so we'll update it just in case.
			update(theGraphics);
		}
	}
	
	/*
	Handle starting and ending games.
	*/
	public void keyPressed(KeyEvent theEvent) {
		if (theEvent.getKeyCode() == KeyEvent.VK_SPACE) {
			if (cGameInProgress)
				endGame();
			else
				newGame();
		}
		
		// let GameApplet know about the game event
		super.keyPressed(theEvent);
	}
	
	public void mousePressed(MouseEvent theEvent) {
		if (cGameInProgress)
			endGame();
		else
			newGame();
	}
	
	public void mouseReleased(MouseEvent theEvent) {}
	public void mouseClicked(MouseEvent theEvent) {}
	public void mouseEntered(MouseEvent theEvent) {}
	public void mouseExited(MouseEvent theEvent) {}
}

