/***************************************************************

Wrapping is primarily intended as a visual representation of a
wrapped buffer. The left view is the buffer as it is normally
shown, the right view shows what goes on "behind the scenes".
The colored rectangles show how the different buffer pieces are
assembled. Wrapping also shows why the buffer size needs to be a
multiple of the tile size.


The source for WrappingApplet is in the public domain. This
means you can use it in your own projects, whole or in part.
Please note that GameForm is not in the public domain, see the
GameForm source code, documentation, or homepage for
information about the allowed and disallowed usage of GameForm.


GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;
import java.net.*;


public class WrappingApplet extends GameApplet implements GLoadListener
{
	// keys
	static final int				kLeft = 0;
	static final int				kRight = 1;
	static final int				kUp = 2;
	static final int				kDown = 3;
	
	static final int				kLevelWidth = 240;
	static final int				kLevelHeight = 320;
	
	
	GLevelCanvas					cCanvas;
	
	// the WrappingPlayer will set these variables, so this WrappingApplet
	// will know where to scroll to center on the player
	int								cTargetCenterX;
	int								cTargetCenterY;
	
	// file loading status
	boolean							cDoneLoading;
	boolean							cFailedLoading;
	
	
	public void init() {
		// prepare GameApplet
		super.init(kDoubleBuffer);
		cFrameLength = 30; // 30 milliseconds, 33 frames per second
		setKeys(5, 3,
				new int[] {KeyEvent.VK_LEFT,
							KeyEvent.VK_RIGHT,
							KeyEvent.VK_UP,
							KeyEvent.VK_DOWN},
				new int[] {kKeyContinous,
							kKeyContinous,
							kKeyContinous,
							kKeyContinous});
		
		// Toggle GLevelCanvas debugging when typing "debug"
		setCheats(new int[][] {{
				KeyEvent.VK_D,
				KeyEvent.VK_E,
				KeyEvent.VK_B,
				KeyEvent.VK_U,
				KeyEvent.VK_G}});
		
		// prepare GameForm
		GFactory.setFactory(new WrappingFactory());
		GOutput.sMute = true;
		
		WrappingPlayer.cApplet = this;
		
		// init buffer
		cCanvas = new GLevelCanvas(this, null, kLevelWidth, kLevelHeight);
		
		// start loading data
		cDoneLoading = false;
		cFailedLoading = false;
		try {
			GFileManager.load(new URL(getCodeBase(), getParameter("file")), this);
		} catch (MalformedURLException e) {
			cFailedLoading = true;
		}
	}
	
	public void pauseGame(boolean pause) {}
	public void endGame() {}
	
	public void newGame() {
		if (!cDoneLoading)
			return;
		
		cCanvas.setLevel(GLevel.get(0));
		cCanvas.setReverseSpriteOrder(true);
		
		super.newGame();
	}
	
	public void cheatActivated(int cheatID) {
		if (cheatID == 0) { // debug
			cCanvas.cDebug = !cCanvas.cDebug;
			/* to remove the gray areas when leaving debugging mode use:
			if (!cCanvas.cDebug)
				cCanvas.setNeedUpdate(true);
			*/
		}
	}
	
	/*
	Move the game state to the next (i.e. current) frame.
	advanceFrame() is synchronized to keep the game thread from changing
	the data while the event thread is updating the screen.
	*/
	public synchronized void advanceFrame() {
		int							targetScrollX;
		int							targetScrollY;
		
		// let GameForm do its thing
		GLevel.get(0).nextFrame(cFrame);
		
		// update scrolling, centering on the player with a slight delay
		targetScrollX = cTargetCenterX - kLevelWidth / 2;
		targetScrollY = cTargetCenterY - kLevelHeight / 2;
		
		cCanvas.setScrollX((cCanvas.getScrollX() * 5 + targetScrollX) / 6);
		cCanvas.setScrollY((cCanvas.getScrollY() * 5 + targetScrollY) / 6);
	}
	
	public void newProgress(int bitsRead) {
		if (bitsRead == GFileManager.kFinished) {
			cDoneLoading = true;
			newGame();
		} else if (bitsRead == GFileManager.kFailed) {
			cFailedLoading = true;
			repaint();
		}
	}
	
	/*
	repaint() has been called.
	*/
	public synchronized void update(Graphics theGraphics) {
		// paint to the buffer
		if (cGameInProgress) {
			Color						topLeft;
			Color						topRight;
			Color						bottomLeft;
			Color						bottomRight;
			
			// a game is running, paint the level to cBuffer
			cBufferGraphics.setColor(Color.white);
			cBufferGraphics.fillRect(0, 0, cBufferWidth, cBufferHeight);
			
			cCanvas.paint(null, cFrame);
			
			cBufferGraphics.drawImage(cCanvas.getBuffer(), 10, 10, null);
			
			cBufferGraphics.drawImage(cCanvas.getPixelBuffer(), kLevelWidth + 20, 10, null);
			
			// draw colored frames...
			if ((cCanvas.getScrollX() / kLevelWidth) % 2 == 0 && (cCanvas.getScrollY() / kLevelHeight) % 2 == 0) {
				topLeft = Color.black;
				topRight = Color.yellow;
				bottomLeft = Color.red;
				bottomRight = new Color(40, 40, 255);
			} else if ((cCanvas.getScrollY() / kLevelHeight) % 2 == 0) {
				topLeft = Color.red;
				topRight = new Color(40, 40, 255);
				bottomLeft = Color.black;
				bottomRight = Color.yellow;
			} else if ((cCanvas.getScrollX() / kLevelWidth) % 2 == 0) {
				topLeft = Color.yellow;
				topRight = Color.black;
				bottomLeft = new Color(40, 40, 255);
				bottomRight = Color.red;
			} else {
				topLeft = new Color(40, 40, 255);
				topRight = Color.red;
				bottomLeft = Color.yellow;
				bottomRight = Color.black;
			}
			
			if ((cCanvas.getScrollY() / kLevelHeight) % 2 == 0 && (cCanvas.getScrollX() / kLevelWidth) % 2 == 0) {
				topLeft = Color.yellow;
				topRight = Color.black;
				bottomLeft = Color.blue;
				bottomRight = Color.red;
			} else if ((cCanvas.getScrollX() / kLevelWidth) % 2 == 0) {
				topLeft = Color.blue;
				topRight = Color.red;
				bottomLeft = Color.yellow;
				bottomRight = Color.black;
			} else if ((cCanvas.getScrollY() / kLevelHeight) % 2 == 0) {
				topLeft = Color.black;
				topRight = Color.yellow;
				bottomLeft = Color.red;
				bottomRight = Color.blue;
			} else {
				topLeft = Color.red;
				topRight = Color.blue;
				bottomLeft = Color.black;
				bottomRight = Color.yellow;
			}
			
			// left view
			cBufferGraphics.setColor(topLeft);
			cBufferGraphics.drawRect(10,
					10,
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 1,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(11,
					11,
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 3,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(topRight);
			cBufferGraphics.drawRect(10 + kLevelWidth - (cCanvas.getScrollX() % kLevelWidth),
					10,
					(cCanvas.getScrollX() % kLevelWidth) - 1,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(11 + kLevelWidth - (cCanvas.getScrollX() % kLevelWidth),
					11,
					(cCanvas.getScrollX() % kLevelWidth) - 3,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(bottomLeft);
			cBufferGraphics.drawRect(10,
					10 + kLevelHeight - (cCanvas.getScrollY() % kLevelHeight),
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 1,
					(cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(11,
					11 + kLevelHeight - (cCanvas.getScrollY() % kLevelHeight),
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 3,
					(cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(bottomRight);
			cBufferGraphics.drawRect(10 + kLevelWidth - (cCanvas.getScrollX() % kLevelWidth),
					10 + kLevelHeight - (cCanvas.getScrollY() % kLevelHeight),
					(cCanvas.getScrollX() % kLevelWidth) - 1,
					(cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(11 + kLevelWidth - (cCanvas.getScrollX() % kLevelWidth),
					11 + kLevelHeight - (cCanvas.getScrollY() % kLevelHeight),
					(cCanvas.getScrollX() % kLevelWidth) - 3,
					(cCanvas.getScrollY() % kLevelHeight) - 3);
			
			// right view
			cBufferGraphics.setColor(topLeft);
			cBufferGraphics.drawRect(kLevelWidth + 20 + (cCanvas.getScrollX() % kLevelWidth),
					10 + (cCanvas.getScrollY() % kLevelHeight),
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 1,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(kLevelWidth + 21 + (cCanvas.getScrollX() % kLevelWidth),
					11 + (cCanvas.getScrollY() % kLevelHeight),
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 3,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(topRight);
			cBufferGraphics.drawRect(kLevelWidth + 20,
					10 + (cCanvas.getScrollY() % kLevelHeight),
					(cCanvas.getScrollX() % kLevelWidth) - 1,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(kLevelWidth + 21,
					11 + (cCanvas.getScrollY() % kLevelHeight),
					(cCanvas.getScrollX() % kLevelWidth) - 3,
					kLevelHeight - (cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(bottomLeft);
			cBufferGraphics.drawRect(20 + kLevelWidth + (cCanvas.getScrollX() % kLevelWidth),
					10,
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 1,
					(cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(21 + kLevelWidth + (cCanvas.getScrollX() % kLevelWidth),
					11,
					kLevelWidth - (cCanvas.getScrollX() % kLevelWidth) - 3,
					(cCanvas.getScrollY() % kLevelHeight) - 3);
			
			cBufferGraphics.setColor(bottomRight);
			cBufferGraphics.drawRect(20 + kLevelWidth,
					10,
					(cCanvas.getScrollX() % kLevelWidth) - 1,
					(cCanvas.getScrollY() % kLevelHeight) - 1);
			cBufferGraphics.drawRect(21 + kLevelWidth,
					11,
					(cCanvas.getScrollX() % kLevelWidth) - 3,
					(cCanvas.getScrollY() % kLevelHeight) - 3);
			
			// copy buffer to screen
			theGraphics.drawImage(cBuffer, 0, 0, null);
		} else {
			// no game is running, display the title screen
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.fillRect(0, 0, getSize().width, getSize().height);
			
			cBufferGraphics.setFont(new Font("SansSerif", Font.PLAIN, 14));
			cBufferGraphics.setColor(Color.white);
			if (cFailedLoading)
				cBufferGraphics.drawString("Failed Loading Data", (getSize().width - cBufferGraphics.getFontMetrics().stringWidth("Failed Loading Data")) / 2, getSize().height / 2 + 10);
			else
				cBufferGraphics.drawString("Loading Data... Please Wait.", (getSize().width - cBufferGraphics.getFontMetrics().stringWidth("Loading Data... Please Wait.")) / 2, getSize().height / 2 + 10);
			// copy buffer to screen
			theGraphics.drawImage(cBuffer, 0, 0, null);
		}
	}
	
	/*
	The view is "damaged"
	*/
	public void paint(Graphics theGraphics) {
		update(theGraphics);
	}
}

